/*******************************************************************************
 *  Copyright (c) 2010 Association for Decentralized Information Management in
 *  Industry THTH ry.
 *  All rights reserved. This program and the accompanying materials
 *  are made available under the terms of the Eclipse Public License v1.0
 *  which accompanies this distribution, and is available at
 *  http://www.eclipse.org/legal/epl-v10.html
 *
 *  Contributors:
 *      VTT Technical Research Centre of Finland - initial API and implementation
 *******************************************************************************/
package org.simantics.databoard.accessor.java;

import org.simantics.databoard.accessor.Accessor;
import org.simantics.databoard.accessor.BooleanAccessor;
import org.simantics.databoard.accessor.error.AccessorConstructionException;
import org.simantics.databoard.accessor.error.AccessorException;
import org.simantics.databoard.accessor.error.ReferenceException;
import org.simantics.databoard.accessor.event.Event;
import org.simantics.databoard.accessor.event.ValueAssigned;
import org.simantics.databoard.accessor.impl.ListenerEntry;
import org.simantics.databoard.accessor.interestset.BooleanInterestSet;
import org.simantics.databoard.accessor.reference.AccessorReference;
import org.simantics.databoard.binding.ArrayBinding;
import org.simantics.databoard.binding.Binding;
import org.simantics.databoard.binding.BooleanBinding;
import org.simantics.databoard.binding.RecordBinding;
import org.simantics.databoard.binding.VariantBinding;
import org.simantics.databoard.binding.error.BindingException;
import org.simantics.databoard.binding.java.BooleanBindingDefault;
import org.simantics.databoard.binding.mutable.Variant;
import org.simantics.databoard.type.BooleanType;

/**
 * Accessor to a Java Object of Boolean Type.
 *
 * @author Toni Kalajainen <toni.kalajainen@vtt.fi>
 */
public class JavaBoolean extends JavaObject implements BooleanAccessor {
	
	public JavaBoolean(JavaObject parent, BooleanBinding binding, Object object) {
		super(parent, binding, object);
	}
	
	public BooleanBinding getBinding() {
		return (BooleanBinding) binding;
	}

	@Override
	public BooleanType type() {
		return getBinding().type();
	}

	@Override
	public boolean getValue() throws AccessorException {
		try {
			return getBinding().getValue_(object);
		} catch (BindingException e) {
			throw new AccessorException(e);
		}
	}

	@Override
	public void setValue(boolean newValue) throws AccessorException {
		try {
			if (binding.isImmutable() && parent!=null && parent instanceof JavaArray) {
				ArrayBinding ab = (ArrayBinding) parent.binding;
				Object nv = getBinding().create(newValue);
				ab.set(parent.object, (Integer)keyInParent, nv);
				this.object = nv;
			} else if (binding.isImmutable() && parent!=null && parent instanceof JavaRecord) {
				RecordBinding rb = (RecordBinding) parent.binding;
				Object nv = getBinding().create(newValue);
				rb.setComponent(parent.object, (Integer)keyInParent, nv);
				this.object = nv;
			} else if (binding.isImmutable() && parent!=null && parent instanceof JavaVariant) {
				VariantBinding vb = (VariantBinding) parent.binding;
				Object nv = getBinding().create(newValue);
				vb.setValue(parent.object, nv, getBinding());
				this.object = nv;
			} else {			
				// Set value
				getBinding().setValue(object, newValue);
			}
			
			// Notify
			ListenerEntry le = listeners;
			while (le!=null) {
				BooleanInterestSet is = le.getInterestSet();
				if (is.inNotifications()) {					
					Event e = new ValueAssigned( BooleanBindingDefault.INSTANCE, is.inValues() ? newValue : null );
					le.emitEvent(e);
				}
				le = le.next;
			}
			
		} catch (BindingException e) {
			throw new AccessorException(e);
		}
	}

	@Override
	public void setValue(Binding binding, Object newValue) throws AccessorException {
		try {
			boolean nv = ((BooleanBinding)binding).getValue_(newValue);
			setValue(nv);
		} catch (BindingException e) {
			throw new AccessorException(e);
		}
	}

	@SuppressWarnings("unchecked")
	public <T extends Accessor> T getAccessor(AccessorReference reference)
	throws AccessorConstructionException {
		if (reference==null) return (T) this;		
		throw new ReferenceException(reference.getClass()+" is not a subreference of BooleanType");	
	}

	@Override
	Event applyLocal(Event e, boolean makeRollback) throws AccessorException {
		Event rollback = null;
		rollback = makeRollback ? new ValueAssigned( BooleanBindingDefault.INSTANCE, getValue() ) : null; 		
		if (e instanceof ValueAssigned) {
			ValueAssigned va = (ValueAssigned) e;
			if (va.newValue == null) throw new AccessorException("Boolean value expected, got null");			
			setValue(va.newValue.getBinding(), va.newValue.getValue());
			return rollback;
		} else throw new AccessorException("Cannot apply "+e.getClass().getName()+" to Boolean");
	}
	
	
}

