/*******************************************************************************
 *  Copyright (c) 2010 Association for Decentralized Information Management in
 *  Industry THTH ry.
 *  All rights reserved. This program and the accompanying materials
 *  are made available under the terms of the Eclipse Public License v1.0
 *  which accompanies this distribution, and is available at
 *  http://www.eclipse.org/legal/epl-v10.html
 *
 *  Contributors:
 *      VTT Technical Research Centre of Finland - initial API and implementation
 *******************************************************************************/
package org.simantics.databoard.accessor.java;

import java.io.IOException;
import java.util.LinkedList;
import java.util.List;

import org.simantics.databoard.Bindings;
import org.simantics.databoard.accessor.Accessor;
import org.simantics.databoard.accessor.error.AccessorConstructionException;
import org.simantics.databoard.accessor.error.AccessorException;
import org.simantics.databoard.accessor.event.Event;
import org.simantics.databoard.accessor.event.InvalidatedEvent;
import org.simantics.databoard.accessor.impl.ListenerEntry;
import org.simantics.databoard.accessor.interestset.InterestSet;
import org.simantics.databoard.accessor.reference.AccessorReference;
import org.simantics.databoard.adapter.AdaptException;
import org.simantics.databoard.binding.ArrayBinding;
import org.simantics.databoard.binding.Binding;
import org.simantics.databoard.binding.BooleanBinding;
import org.simantics.databoard.binding.ByteBinding;
import org.simantics.databoard.binding.DoubleBinding;
import org.simantics.databoard.binding.FloatBinding;
import org.simantics.databoard.binding.IntegerBinding;
import org.simantics.databoard.binding.LongBinding;
import org.simantics.databoard.binding.MapBinding;
import org.simantics.databoard.binding.OptionalBinding;
import org.simantics.databoard.binding.RecordBinding;
import org.simantics.databoard.binding.StringBinding;
import org.simantics.databoard.binding.UnionBinding;
import org.simantics.databoard.binding.VariantBinding;
import org.simantics.databoard.binding.error.BindingException;
import org.simantics.databoard.type.DataType;

/**
 * Accessor to a Java Object.
 * <p>
 * The monitoring contract forbids modifications to the object outside this 
 * accessor object. If you do modifications to the value using other mechanisms,
 * do not use the listening mechanism.
 * <p>
 * Operations cannot be performed safely in multi-threaded environment. The
 * caller must provide external synchronization protection, not just to one 
 * object but the whole accessor tree.  
 *
 * @author Toni Kalajainen <toni.kalajainen@vtt.fi>
 */
public abstract class JavaObject implements Accessor {

	/**  
	 * Strong Reference to the parent. It is needed to keep the parent path
	 * alive as long as its children. This is ensure the children are not 
	 * instantiated more than once.  
	 */
	JavaObject parent;	
	/** The Java object */
	Object object;
	/** Binding */
	Binding binding;
	/** Listeners */
	ListenerEntry listeners = null;
	/** Key in parent, index for fields/arrays, key for maps, */
	Object keyInParent = null;

	public JavaObject(JavaObject parent, Binding binding, Object initialValue) {
		if (binding==null) throw new IllegalArgumentException("null arg");
		this.parent = parent;
		this.binding = binding;
		this.object = initialValue;
	}
	
	/**
	 * Get the Java Object
	 * 
	 * @return Object
	 */
	public Object getObject() {
		return object;
	}

	public Binding getBinding() {
		return binding;
	}	
	
	public DataType type() {
		return binding.type();
	}	

	@Override
	public Object getValue(Binding binding) throws AccessorException {
		try {
			if (binding == this.binding) {
				return binding.isImmutable() ? object : binding.clone(object);
			}
			return Bindings.adapt(object, this.binding, binding);
		} catch (AdaptException e) {
			throw new AccessorException(e);
		}
	}
	
	@Override
	public void addListener(Listener listener, InterestSet interestSet, AccessorReference path) throws AccessorException {
		listeners = ListenerEntry.link(listeners, listener, interestSet, path);
	}

	protected ListenerEntry detachListener(Listener listener) throws AccessorException {
		ListenerEntry e = listeners;
		ListenerEntry p = null;
		while (e!=null) {
			// Found match
			if (e.listener == listener) {
				// The match was the first entry of the linked list
				if (p==null) {
					listeners = e.next;
					return e;
				}
				// Some other entry, unlink e
				p.next = e.next;
				return e;
			}
			p = e;
			e = e.next;
		}
		return null;		
	}
	
	@Override
	public void removeListener(Listener listener) throws AccessorException {
		detachListener(listener);
	}
	
	public static JavaObject createAccessor(Binding b, Object v) throws AccessorConstructionException {
		return createSubAccessor(null, b, v);
	}
	
	static JavaObject createSubAccessor(JavaObject parent, Binding b, Object v) 
	throws AccessorConstructionException {
		if (b instanceof BooleanBinding) {
			return new JavaBoolean(parent, (BooleanBinding)b, v);
		}
		if (b instanceof ByteBinding) {
			return new JavaByte(parent, (ByteBinding)b, v);
		}
		if (b instanceof IntegerBinding) {
			return new JavaInteger(parent, (IntegerBinding)b, v);
		}
		if (b instanceof LongBinding) {
			return new JavaLong(parent, (LongBinding)b, v);
		}
		if (b instanceof FloatBinding) {
			return new JavaFloat(parent, (FloatBinding)b, v);
		}
		if (b instanceof DoubleBinding) {
			return new JavaDouble(parent, (DoubleBinding)b, v);
		}
		if (b instanceof StringBinding) {
			return new JavaString(parent, (StringBinding)b, v);
		}
		if (b instanceof UnionBinding) {
			return new JavaUnion(parent, (UnionBinding)b, v);
		}
		if (b instanceof OptionalBinding) {
			return new JavaOptional(parent, (OptionalBinding)b, v);
		}
		if (b instanceof VariantBinding) {
			return new JavaVariant(parent, (VariantBinding)b, v);
		}
		if (b instanceof ArrayBinding) {
			return new JavaArray(parent, (ArrayBinding)b, v);
		}
		if (b instanceof MapBinding) {
			return new JavaMap(parent, (MapBinding)b, v);
		}
		if (b instanceof RecordBinding) {
			return new JavaRecord(parent, (RecordBinding)b, v);
		}
		throw new AccessorConstructionException("Can not create accessor to "+b.type());
	}
	
	/**
	 * Send notification that this accessor has been detached from the parent
	 */
	void invalidatedNotification() {
		ListenerEntry le = listeners;
		while (le!=null) {			
			InterestSet is = le.getInterestSet();
			if (is.inNotifications()) {
				InvalidatedEvent e = new InvalidatedEvent();
				le.emitEvent(e);
			}
			le = le.next;
		}		
	}

	/**
	 * Apply a change set that has events for the particular accessor. 
	 * There are no sub-accessor events.
	 * 
	 * @param cs
	 * @param makeRollback
	 * @return rollback-event
	 * @throws AccessorException
	 */
	abstract Event applyLocal(Event e, boolean makeRollback) throws AccessorException;
	
	@Override
	public void apply(List<Event> cs, LinkedList<Event> rollback) throws AccessorException {
		try {
			boolean makeRollback = rollback != null;
			for (Event e : cs) {
				// Accessor
				JavaObject a = e.reference == null ? this : (JavaObject) getAccessor(e.reference);
				// Apply changes				
				Event rbe = a.applyLocal(e, makeRollback);
				if (makeRollback) {
					rbe.reference = e.reference;
					rollback.addFirst( rbe );
				}
			}
		} catch (AccessorConstructionException ae) {
			throw new AccessorException(ae);
		}
		
	}
	
	@Override
	public String toString() {
		try {
			return "Java("+binding.printValueDefinition(object, true)+")";
		} catch (IOException e) {
			return "Java(error="+e.getMessage()+")";
		} catch (BindingException e) {
			return "Java(error="+e.getMessage()+")";
		}
	}

}

