/*******************************************************************************
 *  Copyright (c) 2010 Association for Decentralized Information Management in
 *  Industry THTH ry.
 *  All rights reserved. This program and the accompanying materials
 *  are made available under the terms of the Eclipse Public License v1.0
 *  which accompanies this distribution, and is available at
 *  http://www.eclipse.org/legal/epl-v10.html
 *
 *  Contributors:
 *      VTT Technical Research Centre of Finland - initial API and implementation
 *******************************************************************************/
package org.simantics.databoard.accessor.java;

import java.lang.ref.SoftReference;

import org.simantics.databoard.accessor.Accessor;
import org.simantics.databoard.accessor.VariantAccessor;
import org.simantics.databoard.accessor.error.AccessorConstructionException;
import org.simantics.databoard.accessor.error.AccessorException;
import org.simantics.databoard.accessor.event.Event;
import org.simantics.databoard.accessor.event.ValueAssigned;
import org.simantics.databoard.accessor.impl.ListenerEntry;
import org.simantics.databoard.accessor.interestset.InterestSet;
import org.simantics.databoard.accessor.interestset.VariantInterestSet;
import org.simantics.databoard.accessor.reference.AccessorReference;
import org.simantics.databoard.accessor.reference.LabelReference;
import org.simantics.databoard.accessor.reference.VariantValueReference;
import org.simantics.databoard.adapter.AdaptException;
import org.simantics.databoard.binding.Binding;
import org.simantics.databoard.binding.VariantBinding;
import org.simantics.databoard.binding.error.BindingException;
import org.simantics.databoard.binding.mutable.ValueContainer;
import org.simantics.databoard.binding.mutable.Variant;
import org.simantics.databoard.binding.mutable.VariantContainerBinding;
import org.simantics.databoard.type.DataType;
import org.simantics.databoard.type.VariantType;

public class JavaVariant extends JavaObject implements VariantAccessor {

	SoftReference<JavaObject> child;	
	
	public JavaVariant(JavaObject parent, VariantBinding binding, Object object) {
		super(parent, binding, object);
	}
	
	@Override
	public VariantType type() {
		return (VariantType) binding.type();
	}
	
	@Override
	public VariantBinding getBinding() {
		return (VariantBinding) binding;
	}

	@SuppressWarnings("unchecked")
	@Override
	public <T extends Accessor> T getValueAccessor() throws AccessorConstructionException {
		try {
			JavaObject sa = getExistingAccessor();
			if (sa==null) {
				// Create new child
				Binding cb = getBinding().getBinding(object);
				Object cv = getBinding().getValue(object);
				sa = JavaObject.createSubAccessor(this, cb, cv);
				child = new SoftReference<JavaObject>(sa);
				
				// Attach listeners of interest set to the new accessor (of the value)
				ListenerEntry le = listeners;
				while (le!=null) {				
					VariantInterestSet is = le.getInterestSet();
					InterestSet cis = is.getComponentInterest();
					if (cis != null) {
						try {
							AccessorReference childPath = AccessorReference.concatenate(le.path, new VariantValueReference() );
							sa.addListener(le.listener, cis, childPath);
						} catch (AccessorException e) {
							throw new AccessorConstructionException(e);
						}
					}
					if (is.inCompleteComponent()) {
						cis = InterestSet.newInterestSet(getContentType(), true, true, true);
						try {
							AccessorReference childPath = AccessorReference.concatenate(le.path, new VariantValueReference() );
							sa.addListener(le.listener, cis, childPath);
						} catch (AccessorException e) {
							throw new AccessorConstructionException(e);
						}
					}
					le = le.next;
				}				
				
			}
			return (T) sa;
		} catch (BindingException e) {
			throw new AccessorConstructionException(e);
		} catch (AccessorException e) {
			throw new AccessorConstructionException(e);
		}
	}

	@SuppressWarnings("unchecked")
	@Override
	public <T extends Accessor> T getAccessor(AccessorReference reference)
			throws AccessorConstructionException {		
		if (reference==null) return (T) this;
		
		if (reference instanceof LabelReference) {
			LabelReference lr = (LabelReference) reference;
			if (lr.label.equals("v")) {
				Accessor sa = getValueAccessor();
				if (reference.getChildReference()!=null) sa = sa.getAccessor(reference.getChildReference());
				return (T) sa;
			}			
		}		
		
		if (reference instanceof VariantValueReference) {
			Accessor sa = getValueAccessor();
			if (reference.getChildReference()!=null) sa = sa.getAccessor(reference.getChildReference());
			return (T) sa;
		}
		
		throw new AccessorConstructionException("Variant value reference expected, got "+reference.getClass().getName());
	}

	protected JavaObject getExistingAccessor() {
		SoftReference<JavaObject> r = child;
		if (r==null) return null;
		JavaObject sa = r.get();
		return sa;
	}
	
	@Override
	public void setValue(Binding variantBinding, Object newVariant)
		throws AccessorException {
		try {
			if (binding instanceof VariantBinding == false)
				throw new AccessorException("VariantBinding is expected.");
			
			Binding newValueBinding = ((VariantBinding)variantBinding).getBinding(newVariant);
			Object newValueObject = ((VariantBinding)variantBinding).getValue(newVariant);
			setContentValue(newValueBinding, newValueObject);
		} catch (BindingException e) {
			throw new AccessorException( e ); 
		}
	}

	@Override
	public Object getContentValue(Binding contentBinding)
	throws AccessorException {
		try {
			return getBinding().getValue(object);
		} catch (BindingException e) {
			throw new AccessorException(e);
		}
	}
	
	@Override
	public DataType getContentType() throws AccessorException {
		try {
			return getBinding().getType(object);
		} catch (BindingException e) {
			throw new AccessorException(e);
		}
	}
	
	@Override
	public void setContentValue(Binding valueBinding, Object newValue)
	throws AccessorException {
		try {
			// reuse sub-accessor, if type matches, and sub-accessor exists
			JavaObject sa = getExistingAccessor();
			
			if (sa!=null && valueBinding.type().equals(sa.type())) {
				sa.setValue(valueBinding, newValue);
				return;
			}
			
			{
				// Replace the old value with a new value
				if (sa!=null) {
					sa.invalidatedNotification();
					child = null;
					sa = null;
				}
				
				// Write new value
				getBinding().setValue(object, newValue, valueBinding);
				
				// Notify Listeners
				ListenerEntry le = listeners;
				while (le!=null) {
					// Notification
					VariantInterestSet is = le.getInterestSet();
					if (is.inNotifications()) {
						if (is.inValues()) {
							Variant value = new Variant( valueBinding, valueBinding.isImmutable() ? newValue : valueBinding.clone(newValue) );
							ValueAssigned e = new ValueAssigned(VariantContainerBinding.INSTANCE, value );
							le.emitEvent(e);
						} else {
							le.emitEvent( new ValueAssigned() );							
						}
					}
					
					// Attach component listener
//					InterestSet cis = is.getComponentInterest();
//					if (cis!=null) {
//						sa = getValueAccessor();
//					}
					
					le = le.next;
				}
				
			}
		} catch (BindingException e) {
			throw new AccessorException(e);
//		} catch (AccessorConstructionException e) {
//			throw new AccessorException(e);
		} catch (AdaptException e) {
			throw new AccessorException(e);
		}
	}
	
	@Override
	public void addListener(Listener listener, InterestSet interestSet,
			AccessorReference path) throws AccessorException {
		super.addListener(listener, interestSet, path);
		VariantInterestSet is = (VariantInterestSet) interestSet;
		InterestSet cis = is.getComponentInterest();
		if (cis==null && !is.inCompleteComponent()) return;
		JavaObject sa = getExistingAccessor();
		if (sa==null) return;
		if (cis!=null) {
			AccessorReference childPath = AccessorReference.concatenate(path, new VariantValueReference() );
			sa.addListener(listener, cis, childPath);
		}
		if (is.inCompleteComponent()) {
			AccessorReference childPath = AccessorReference.concatenate(path, new VariantValueReference() );
			cis = InterestSet.newInterestSet(getContentType(), true, true, true);
			sa.addListener(listener, cis, childPath);
		}
	}		
	
	@Override
	public void removeListener(Listener listener) throws AccessorException {
		ListenerEntry e = detachListener(listener);
		if (e==null) return;
		VariantInterestSet is = (VariantInterestSet) e.interestSet;
		
		InterestSet cis = is.getComponentInterest();
		if (cis==null && !is.inCompleteComponent()) return;
		JavaObject sa = getExistingAccessor();
		if (sa==null) return;
		if (cis!=null) {
			sa.removeListener(listener);
		}
		// Remove another?
		if (is.inCompleteComponent()) {
			sa.removeListener(listener);
		}		
	}


	@Override
	Event applyLocal(Event e, boolean makeRollback) throws AccessorException {
		try {
			Event rollback = null;			
			if (e instanceof ValueAssigned) {
				ValueAssigned va = (ValueAssigned) e; 
				if (va.newValue==null) throw new AccessorException("Cannot apply variant assignment event, the value is missing");
				
				if (makeRollback) {
					Binding cb = getBinding().getBinding(object);
					Object cv = getBinding().getValue(object);
					rollback = new ValueAssigned( VariantContainerBinding.INSTANCE, new Variant( cb, cv ) );
				}
	
				setValue(va.newValue.getBinding(), va.newValue.getValue());
//				setContentValue(va.newValue.getBinding(), va.newValue.getValue());
				
				return rollback;
			} else {
				throw new AccessorException("Invalid event "+e.getClass().getName());
			}
		} catch (BindingException be) {
			throw new AccessorException(be);
		}
	}
	
}

