/*******************************************************************************
 *  Copyright (c) 2010 Association for Decentralized Information Management in
 *  Industry THTH ry.
 *  All rights reserved. This program and the accompanying materials
 *  are made available under the terms of the Eclipse Public License v1.0
 *  which accompanies this distribution, and is available at
 *  http://www.eclipse.org/legal/epl-v10.html
 *
 *  Contributors:
 *      VTT Technical Research Centre of Finland - initial API and implementation
 *******************************************************************************/
package org.simantics.databoard.binding;

import java.util.IdentityHashMap;
import java.util.Set;

import org.simantics.databoard.Bindings;
import org.simantics.databoard.binding.error.BindingException;
import org.simantics.databoard.binding.error.RuntimeBindingException;
import org.simantics.databoard.binding.mutable.Variant;
import org.simantics.databoard.binding.mutable.VariantContainerBinding;
import org.simantics.databoard.reflection.ObjectVariantBinding;
import org.simantics.databoard.type.DataType;
import org.simantics.databoard.type.VariantType;
import org.simantics.databoard.util.DataValueUtil;
import org.simantics.databoard.util.IdentityPair;

/**
 * This is an abstract binding of VariantType to a Java Object.
 * 
 * Variant is a container for datavalue of any Datatype. 
 * 
 * @see VariantType The Datatype
 * @see VariantContainerBinding Binds variant to {@link Variant}
 * @see ObjectVariantBinding Binds variant to java.lang.Object
 * @see StringVariantBindingExample Binds variant to java.lang.String (Filename and URL compatible) 
 * @author Toni Kalajainen
 */
public abstract class VariantBinding extends Binding {
	
	public VariantBinding()
	{
		super();
		this.type = VariantType.INSTANCE;
	}

	/**
	 * Get the value in the variant. The returned value may represent internal
	 * value of <code>variant</code>.  
	 * 
	 * @param variant
	 * @param binding
	 * @return value
	 * @throws BindingException
	 */
	public abstract Object getValue(Object variant, Binding binding)
	throws BindingException;

	/**
	 * Get the value of the variant. The value is bound with the suggested
	 * binding, see {@link #getBinding(Object)}. 
	 * 
	 * @param variant
	 * @return value
	 * @throws BindingException
	 */
	public abstract Object getValue(Object variant)
	throws BindingException;
	
	public abstract DataType getType(Object variant)
	throws BindingException;

	/**
	 * Create a new variant object. The <code>value</code> argument may be included in the result. 
	 * 
	 * @param value
	 * @param binding
	 * @return variant
	 * @throws BindingException
	 */
	public abstract Object create(Object value, Binding binding)
	throws BindingException;
	
	public Object createUnchecked(Object value, Binding binding)
	throws RuntimeBindingException {
		try {
			return create(value, binding);
		} catch (BindingException e) {
			throw new RuntimeBindingException(e);
		}
	}
	
	/**
	 * Returns a suggestion for the binding of the value of the variant.
	 * The binding is a suggestion for most optimal usage.  
	 * 
	 * @param variant
	 * @return binding
	 */
	public abstract Binding getBinding(Object variant)
	throws BindingException;
	
	/**
	 * Set the value of an existing variant container.
	 *  
	 * Note, This is not implemented in all bindings.   
	 * 
	 * @param variant
	 * @param value
	 * @param binding
	 * @throws BindingException
	 */
	public abstract void setValue(Object variant, Object value, Binding binding)
	throws BindingException;	
		
    @Override
    public void accept(Visitor1 v, Object obj) {
        v.visit(this, obj);        
    }
    
    @Override
    public <T> T accept(Visitor<T> v) {
        return v.visit(this);
    }

    @Override
    public int deepHashValue(Object value, IdentityHashMap<Object, Object> hashedObjects) throws BindingException {
   		DataType type = getType(value);    	
   		Binding binding = getBinding(value);
		Binding dataTypeBinding = Bindings.getBindingUnchecked(DataType.class);
    	Object element = getValue(value, binding);
    	return dataTypeBinding.deepHashValue(type, hashedObjects) + binding.deepHashValue(element, hashedObjects);
    }

	@Override
	public int deepCompare(Object o1, Object o2,
			Set<IdentityPair<Object, Object>> compareHistory)
			throws BindingException {
		// Compare Type
		DataType t1 = getType(o1);
		DataType t2 = getType(o2);
		Binding dataTypeBinding = Bindings.getBindingUnchecked(DataType.class);
		int dif = dataTypeBinding.compare(t1, t2);
		if (dif!=0) return dif;
		// Compare Value
		Binding bi1 = getBinding(o1);
		Binding bi2 = getBinding(o2);
		Object va1 = getValue(o1, bi1);
		Object va2 = getValue(o2, bi2);
		return DataValueUtil.compare(bi1, va1, bi2, va2);		
	}
	
}

