/*******************************************************************************
 *  Copyright (c) 2010 Association for Decentralized Information Management in
 *  Industry THTH ry.
 *  All rights reserved. This program and the accompanying materials
 *  are made available under the terms of the Eclipse Public License v1.0
 *  which accompanies this distribution, and is available at
 *  http://www.eclipse.org/legal/epl-v10.html
 *
 *  Contributors:
 *      VTT Technical Research Centre of Finland - initial API and implementation
 *******************************************************************************/
package org.simantics.databoard.binding.mutable;

import java.io.IOException;
import java.io.InvalidObjectException;
import java.io.ObjectStreamException;
import java.io.Serializable;

import org.simantics.databoard.Bindings;
import org.simantics.databoard.adapter.AdaptException;
import org.simantics.databoard.binding.Binding;
import org.simantics.databoard.binding.error.BindingException;
import org.simantics.databoard.binding.error.RuntimeBindingException;
import org.simantics.databoard.parser.DataValuePrinter;
import org.simantics.databoard.parser.PrintFormat;
import org.simantics.databoard.parser.repository.DataValueRepository;
import org.simantics.databoard.parser.unparsing.DataTypePrinter;
import org.simantics.databoard.reflection.VoidBinding;
import org.simantics.databoard.serialization.SerializationException;
import org.simantics.databoard.type.DataType;

/**
 * Variant is a container to a data value of any type. 
 * The value and type can be changed. 
 *
 * Variant is hash-equals-comparable, even variants of bindings (and types). 
 * The hash function and comparison rules are defined in the manual.   
 *
 * @see VariantContainerBinding is binding for Variant-class
 * @author Toni Kalajainen <toni.kalajainen@vtt.fi>
 */
public class Variant implements Comparable<Variant>, Serializable, Cloneable {

	private static final long serialVersionUID = 1L;
	
	public static Variant ofInstance(Object instance) {
		Binding binding = Bindings.getBindingUnchecked( instance.getClass() );
		return new Variant(binding, instance);
	}
	
	Binding binding;
	Object value;
	
	/**
	 * Constract a variant with a default value of empty record {}
	 */
	public Variant() {
		binding = VoidBinding.VOID_BINDING;
		value = null;
	}
	
	public Variant(Binding binding, Object value) {
		assert(isValid(binding, value));
		this.binding = binding;
		this.value = value;
	}
	
	public Object getValue() {
		return value;
	}
	
	public Object getValue(Binding binding) throws AdaptException {
		return Bindings.adapt(value, this.binding, binding);
	}
	
	public DataType type() {
		return binding.type();
	}
	
	public Binding getBinding() {
		return binding;
	}
	
	public void setValue(Binding binding, Object newValue) {
		assert(isValid(binding, newValue));
		this.binding = binding;
		this.value = newValue;
	}
	
	@Override
	public int hashCode() {
		try {
			return binding.hashValue(value);
		} catch (BindingException e) {
			throw new RuntimeBindingException(e);
		}
	}
	
	@Override
	public boolean equals(Object obj) {
		if(this == obj)
			return true;
		if(obj == null)
			return false;
		try {			
			Variant o = (Variant) obj;
			return Bindings.equals(binding, value, o.binding, o.value);
		} catch (BindingException e) {
			throw new RuntimeBindingException(e);
		}
	}
	
	@Override
	public int compareTo(Variant o) {
		try {
			return Bindings.compare(binding, value, o.binding, o.value);
		} catch (BindingException e) {
			throw new RuntimeBindingException(e);
		}
	}
	
	@Override
	public String toString() {
		try {
			StringBuilder sb = new StringBuilder();			
			DataValueRepository repo = new DataValueRepository();
			DataValuePrinter printer = new DataValuePrinter(sb, repo);
			printer.setFormat(PrintFormat.SINGLE_LINE);
			printer.print(binding, value);
			sb.append(" : ");
			DataTypePrinter printer2 = new DataTypePrinter(sb);
			printer2.print(binding.type());
			return sb.toString();
		} catch (IOException e) {
			throw new RuntimeException(e);
		} catch (BindingException e) {
			throw new RuntimeBindingException(e);
		}
	}
	
	public Variant clone() {
		if (binding.isImmutable()) return new Variant(binding, value);
		Object newValue = Bindings.cloneUnchecked(value, binding, binding);
		return new Variant(binding, newValue);
	}
	
	private void writeObject(java.io.ObjectOutputStream out) throws IOException {
		try {
			Binding dataTypeBinding = Bindings.getBindingUnchecked( DataType.class );
			dataTypeBinding.serializer().serialize(type());
			binding.serializer().serialize(value, out);
		} catch (SerializationException e) {
			throw new IOException(e);
		} catch (BindingException e) {
			throw new RuntimeBindingException(e);
		}
	}
	
	private void readObject(java.io.ObjectInputStream in) throws IOException, ClassNotFoundException {
		try {
			Binding dataTypeBinding = Bindings.getBindingUnchecked( DataType.class );
			DataType type = (DataType) dataTypeBinding.serializer().deserialize(in);
			Binding binding = Bindings.getBinding(type);
			Object value = binding.serializer().deserialize(in);
			this.binding = binding;
			this.value = value;
		} catch (BindingException e) {
			throw new RuntimeBindingException(e);
		} catch (SerializationException e) {
			throw new IOException(e);
		}
	}
	
	@SuppressWarnings("unused")
	private void readObjectNoData() throws ObjectStreamException {
		throw new InvalidObjectException("Don't know how to instantiate "+type()+" with no data");		
	}

	boolean isValid(Binding binding, Object obj) throws RuntimeBindingException {
		try {
			binding.assertInstaceIsValid(obj);
			return true;
		} catch (BindingException e) {
			throw new RuntimeBindingException(e);
		}
	}
	
}

