/*******************************************************************************
 *  Copyright (c) 2010 Association for Decentralized Information Management in
 *  Industry THTH ry.
 *  All rights reserved. This program and the accompanying materials
 *  are made available under the terms of the Eclipse Public License v1.0
 *  which accompanies this distribution, and is available at
 *  http://www.eclipse.org/legal/epl-v10.html
 *
 *  Contributors:
 *      VTT Technical Research Centre of Finland - initial API and implementation
 *******************************************************************************/
package org.simantics.databoard.bindingscheme;

import java.lang.ref.SoftReference;
import java.util.Map;
import java.util.WeakHashMap;

import org.simantics.databoard.binding.Binding;
import org.simantics.databoard.binding.RecordBinding;
import org.simantics.databoard.binding.UnionBinding;
import org.simantics.databoard.binding.error.BindingException;
import org.simantics.databoard.binding.java.ArrayListBinding;
import org.simantics.databoard.binding.java.TreeMapBinding;
import org.simantics.databoard.binding.mutable.ContainerOptionalBinding;
import org.simantics.databoard.binding.mutable.MutableBooleanBinding;
import org.simantics.databoard.binding.mutable.MutableByteBinding;
import org.simantics.databoard.binding.mutable.MutableDoubleBinding;
import org.simantics.databoard.binding.mutable.MutableFloatBinding;
import org.simantics.databoard.binding.mutable.MutableIntegerBinding;
import org.simantics.databoard.binding.mutable.MutableLongBinding;
import org.simantics.databoard.binding.mutable.MutableStringBinding;
import org.simantics.databoard.binding.mutable.TaggedObject;
import org.simantics.databoard.binding.mutable.VariantContainerBinding;
import org.simantics.databoard.type.ArrayType;
import org.simantics.databoard.type.BooleanType;
import org.simantics.databoard.type.ByteType;
import org.simantics.databoard.type.DataType;
import org.simantics.databoard.type.DoubleType;
import org.simantics.databoard.type.FloatType;
import org.simantics.databoard.type.IntegerType;
import org.simantics.databoard.type.LongType;
import org.simantics.databoard.type.MapType;
import org.simantics.databoard.type.OptionalType;
import org.simantics.databoard.type.RecordType;
import org.simantics.databoard.type.StringType;
import org.simantics.databoard.type.UnionType;
import org.simantics.databoard.type.VariantType;

/**
 * GenericBinding is a binding that binds any DataType to an Object.
 * Generinc binding binds to completely mutable java classes. 
 * 
 * DataType           | Class of the bound instance
 * ===================|==================
 * BooleanType        | MutableBoolean.class
 * ByteType           | MutableByte.class
 * FloatType          | MutableFloat.class
 * DoubleType         | MutableDouble.class
 * IntegerType        | MutableInt.class
 * LongType           | MutableLong.class
 * StringType         | MutableString.class
 * UnionType          | GenericBinding.TaggedObject.class
 * OptionType         | ValueContainer.class
 * RecordType         | Object[].class
 * ArrayType          | ArrayList.class
 * MapType            | TreeMap.class
 * VariantType        | Variant.class
 *
 * @author Toni Kalajainen <toni.kalajainen@vtt.fi>
 */
public class GenericBindingScheme implements BindingScheme {
	
	Map<DataType, SoftReference<Binding>> cache = new WeakHashMap<DataType, SoftReference<Binding>>();

	@Override
	public synchronized Binding getBinding(DataType type) 
	{
		Binding result = getCachedBinding(type);
		if (result==null) {
			result = createBindingUnsynchronized(type);
			Binding r2 = getCachedBinding(type);
			if (r2==null) {
				cache(type, result);
			}
		}
		return result;
	}		
	
	private Binding getCachedBinding(DataType type) 
	{
		SoftReference<Binding> ref = cache.get(type);
		return ref==null ? null : ref.get();
	}			
	
	private void cache(DataType type, Binding binding) {
		cache.put(type, new SoftReference<Binding>(binding));
	}
	
	@Override
	public boolean supportsType(DataType type) {
		return true;
	}
	
	/**
	 * Create binding to a type
	 * 
	 * @param type data type
	 * @return Binding
	 */
	private Binding createBindingUnsynchronized(DataType type)
	{
		Binding cachedResult = getCachedBinding(type);
		if (cachedResult!=null) return cachedResult;
		
		if (type instanceof BooleanType) return new MutableBooleanBinding((BooleanType)type);
		if (type instanceof DoubleType) return new MutableDoubleBinding((DoubleType)type);
		if (type instanceof FloatType) return new MutableFloatBinding((FloatType)type);
		if (type instanceof ByteType) return new MutableByteBinding((ByteType)type);
		if (type instanceof IntegerType) return new MutableIntegerBinding((IntegerType)type);
		if (type instanceof LongType) return new MutableLongBinding((LongType)type);
		if (type instanceof StringType) return new MutableStringBinding((StringType)type);
//		if (type instanceof VariantType) return VariantType.INSTANCE;
//		if (type instanceof EntityType) return EntityType.
		if (type instanceof ArrayType) {
			ArrayType arrayType = (ArrayType) type;
			DataType componentType = arrayType.getComponentType();
			/*
			if (arrayType.equals(BooleanArrayBinding.INSTANCE.getDataType()) ) return BooleanArrayBinding.INSTANCE;
			if (arrayType.equals(DoubleArrayBinding.INSTANCE.getDataType()) ) return DoubleArrayBinding.INSTANCE;
			if (arrayType.equals(FloatArrayBinding.INSTANCE.getDataType()) ) return FloatArrayBinding.INSTANCE;
			if (arrayType.equals(IntArrayBinding.INSTANCE.getDataType()) ) return IntArrayBinding.INSTANCE;
			if (arrayType.equals(ByteArrayBinding.INSTANCE.getDataType()) ) return ByteArrayBinding.INSTANCE;
			if (arrayType.equals(LongArrayBinding.INSTANCE.getDataType()) ) return LongArrayBinding.INSTANCE;
			return new ObjectArrayBinding( (ArrayType) type, createBindingUnsynchronized(componentType) );
			*/
			return new ArrayListBinding( arrayType, createBindingUnsynchronized(componentType) );
		}
		
		if (type instanceof OptionalType) {
			OptionalType optionalType = (OptionalType) type;
			DataType componentType = optionalType.componentType;
			ContainerOptionalBinding result = new ContainerOptionalBinding( optionalType, null );
			cache(type, result);
			result.componentBinding = createBindingUnsynchronized(componentType);			
			return result;
			
		}
		
		if (type instanceof RecordType) {			
			RecordType recordType = (RecordType) type;
			Binding componentBindings[] = new Binding[ recordType.count() ];
			GenericRecordBinding result = new GenericRecordBinding(recordType, componentBindings);
			cache(type, result);
			result.postConstruction();
			return result;
		}
		
		if (type instanceof UnionType) {
			UnionType unionType = (UnionType) type;
			Binding componentBindings[] = new Binding[ unionType.components.length ];
			GenericUnionBinding result = new GenericUnionBinding(unionType, componentBindings);
			cache(type, result);
			result.postConstruction();
			return result;
		}		
		
		if (type instanceof MapType) {			
			MapType mapType = (MapType) type;
			TreeMapBinding result = new TreeMapBinding(mapType, null, null);
//			HashMapBinding result = new HashMapBinding(mapType, null, null);
			cache(type, result);
			Binding keyBinding = getBinding(mapType.getKeyType());
			Binding valueBinding = getBinding(mapType.getValueType());
			result.setKeyBinding(keyBinding);
			result.setValueBinding(valueBinding);
			result.postConstruction();
			return result;
		}
		
		if (type instanceof VariantType) {			
			VariantContainerBinding result = VariantContainerBinding.INSTANCE;
			cache(type, result);
			return result;
		}
		
		return null;
	}

/**
 * Binds RecordType to Object[]
 *
 * @author Toni Kalajainen <toni.kalajainen@vtt.fi>
 */
class GenericRecordBinding extends RecordBinding {

	public GenericRecordBinding(RecordType type, Binding[] componentBindings)
	{
		this.componentBindings = componentBindings;
        if (type==null) throw new IllegalArgumentException("null arg");
		this.type = type;
	}
	
	public void postConstruction() 
	{
		for (int i=0; i<componentBindings.length; i++)
		{
			DataType componentType = ((RecordType)type).getComponentType(i);
			componentBindings[ i ] = createBindingUnsynchronized( componentType );
		}		
	}
	
	@Override
	public Object create(Object... value) {		
//		Object[] result = new Object[ value.length ];
//		System.arraycopy(value, 0, result, 0, value.length);
//		return result;
		return value;
	}
	
	@Override
	public Object createPartial() {
		return new Object[count()];
	}
	
	@Override
	public Object getComponent(Object obj, int index) {
		return ((Object[])obj)[index];
	}

	@Override
	public void setComponents(Object obj, Object... value) throws BindingException {
		System.arraycopy(value, 0, obj, 0, value.length);
	}
	
	@Override
	public void setComponent(Object obj, int index, Object value)
			throws BindingException {
		Object[] array = (Object[]) obj;
		array[index] = value;
	}

	@Override
	public boolean isInstance(Object obj) {
		return (obj instanceof Object[]) && (((Object[])obj).length == componentBindings.length);
	}
	
}

/**
 * Binds UnionType to TaggedObject.class
 *
 * @author Toni Kalajainen <toni.kalajainen@vtt.fi>
 */
class GenericUnionBinding extends UnionBinding {

	public GenericUnionBinding(UnionType type, Binding[] componentBindings)
	{
		this.componentBindings = componentBindings;
        if (type==null) throw new IllegalArgumentException("null arg");
		this.type = type;
	}

	public void postConstruction() 
	{
		for (int i=0; i<componentBindings.length; i++)
		{
			DataType componentType = ((UnionType)type).components[i].type;
			componentBindings[ i ] = createBindingUnsynchronized( componentType );
		}		
	}

	@Override
	public Object create(int tag, Object value) {
		TaggedObject result = new TaggedObject();
		result.tag = tag;
		result.value = value;
		return result;
	}
	
	@Override
	public void setValue(Object union, int tag, Object value)
	throws BindingException {
		TaggedObject result = (TaggedObject)union;
		result.tag = tag;
		result.value = value;
	}

	@Override
	public int getTag(Object obj) throws BindingException {
		if (!isInstance(obj)) throw new BindingException("Unexpected class "+obj.getClass().getSimpleName()+", TaggedObject expected");
		TaggedObject to = (TaggedObject) obj;
		return to.tag;
	}

	@Override
	public Object getValue(Object obj) throws BindingException {
		if (!isInstance(obj)) throw new BindingException("Unexpected class "+obj.getClass().getSimpleName()+", TaggedObject expected");
		TaggedObject to = (TaggedObject) obj;
		return to.value;
	}

	@Override
	public boolean isInstance(Object obj) {
		return obj instanceof TaggedObject;
	}
	
}

}



