/*******************************************************************************
 *  Copyright (c) 2010 Association for Decentralized Information Management in
 *  Industry THTH ry.
 *  All rights reserved. This program and the accompanying materials
 *  are made available under the terms of the Eclipse Public License v1.0
 *  which accompanies this distribution, and is available at
 *  http://www.eclipse.org/legal/epl-v10.html
 *
 *  Contributors:
 *      VTT Technical Research Centre of Finland - initial API and implementation
 *******************************************************************************/
package org.simantics.databoard.bindingscheme;

import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.simantics.databoard.binding.ArrayBinding;
import org.simantics.databoard.binding.Binding;
import org.simantics.databoard.binding.MapBinding;
import org.simantics.databoard.binding.OptionalBinding;
import org.simantics.databoard.binding.RecordBinding;
import org.simantics.databoard.binding.UnionBinding;
import org.simantics.databoard.binding.VariantBinding;
import org.simantics.databoard.binding.error.BindingException;
import org.simantics.databoard.binding.error.UnsupportedOperationException;
import org.simantics.databoard.binding.java.BooleanArrayBinding;
import org.simantics.databoard.binding.java.BooleanBindingDefault;
import org.simantics.databoard.binding.java.ByteArrayBinding;
import org.simantics.databoard.binding.java.ByteBindingDefault;
import org.simantics.databoard.binding.java.DoubleArrayBinding;
import org.simantics.databoard.binding.java.DoubleBindingDefault;
import org.simantics.databoard.binding.java.FloatArrayBinding;
import org.simantics.databoard.binding.java.FloatBindingDefault;
import org.simantics.databoard.binding.java.IntArrayBinding;
import org.simantics.databoard.binding.java.IntegerBindingDefault;
import org.simantics.databoard.binding.java.LongArrayBinding;
import org.simantics.databoard.binding.java.LongBindingDefault;
import org.simantics.databoard.binding.java.StringBindingDefault;
import org.simantics.databoard.serialization.Serializer;
import org.simantics.databoard.type.ArrayType;
import org.simantics.databoard.type.BooleanType;
import org.simantics.databoard.type.ByteType;
import org.simantics.databoard.type.DataType;
import org.simantics.databoard.type.DoubleType;
import org.simantics.databoard.type.FloatType;
import org.simantics.databoard.type.IntegerType;
import org.simantics.databoard.type.LongType;
import org.simantics.databoard.type.MapType;
import org.simantics.databoard.type.OptionalType;
import org.simantics.databoard.type.RecordType;
import org.simantics.databoard.type.StringType;
import org.simantics.databoard.type.UnionType;
import org.simantics.databoard.type.VariantType;
import org.simantics.databoard.util.IdentityPair;

/**
 * NoBinding is a fake binding that is used with {@link Serializer} to validate 
 * the format of the data with out instantiating Objects. 
 *
 * @author Toni Kalajainen <toni.kalajainen@vtt.fi>
 */
public class UnboundScheme implements BindingScheme {
	
	Map<DataType, Binding> cache = new HashMap<DataType, Binding>();

	@Override
	public synchronized Binding getBinding(DataType type) {
		Binding result = cache.get(type);
		if (result==null) {
			result = createBindingUnsynchronized(type);
			cache.put(type, result);
		}
		return result;
	}		
	
	@Override
	public boolean supportsType(DataType type) {
		return true;
	}	
	
	/**
	 * Create binding to a type
	 * 
	 * @param type data type
	 * @return Binding
	 */
	private Binding createBindingUnsynchronized(DataType type)
	{
		if (cache.containsKey(type)) return cache.get(type);
		
		if (type instanceof BooleanType) return BooleanBindingDefault.INSTANCE;
		if (type instanceof DoubleType) return DoubleBindingDefault.INSTANCE;
		if (type instanceof FloatType) return FloatBindingDefault.INSTANCE;
		if (type instanceof ByteType) return ByteBindingDefault.INSTANCE;
		if (type instanceof IntegerType) return IntegerBindingDefault.INSTANCE;
		if (type instanceof LongType) return LongBindingDefault.INSTANCE;
		if (type instanceof StringType) return StringBindingDefault.INSTANCE;
		if (type instanceof ArrayType) {
			ArrayType arrayType = (ArrayType) type;
			DataType componentType = arrayType.componentType;
			if (componentType instanceof BooleanType) return BooleanArrayBinding.INSTANCE;
			if (componentType instanceof DoubleType) return DoubleArrayBinding.INSTANCE;
			if (componentType instanceof FloatType) return FloatArrayBinding.INSTANCE;
			if (componentType instanceof IntegerType) return IntArrayBinding.INSTANCE;
			if (componentType instanceof ByteType) return ByteArrayBinding.INSTANCE;
			if (componentType instanceof LongType) return LongArrayBinding.INSTANCE;	
			return new UnboundArray( arrayType, createBindingUnsynchronized(componentType) );
		}

		if (type instanceof MapType) {
			MapType mapType = (MapType) type;
			DataType keyType = mapType.getKeyType();
			DataType valueType = mapType.getValueType();
			Binding keyBinding = createBindingUnsynchronized(keyType);
			Binding valueBinding = createBindingUnsynchronized(valueType);
			return new UnboundMap( mapType, keyBinding, valueBinding );
		}

		if (type instanceof VariantType) {
//			VariantType variantType = (VariantType) type;
			return new UnboundVariant();
		}
		
		if (type instanceof OptionalType) {
			OptionalType optionalType = (OptionalType) type;
			DataType componentType = optionalType.componentType;
			UnboundOptionalBinding result = new UnboundOptionalBinding( optionalType );
			cache.put(type, result);
			result.componentBinding = createBindingUnsynchronized(componentType);
			return result;
		}
		
		if (type instanceof RecordType) {			
			RecordType recordType = (RecordType) type;
			Binding componentBindings[] = new Binding[ recordType.count() ];
			UnboundRecord result = new UnboundRecord(recordType, componentBindings);
			cache.put(type, result);
			result.postConstruction();
			return result;
		}
		
		if (type instanceof UnionType) {
			UnionType unionType = (UnionType) type;
			Binding componentBindings[] = new Binding[ unionType.components.length ];
			UnboundUnion result = new UnboundUnion(unionType, componentBindings);
			cache.put(type, result);
			result.postConstruction();
			return result;
		}		
		
		return null;
	}
	

class UnboundRecord extends RecordBinding {

	public UnboundRecord(RecordType type, Binding[] componentBindings)
	{
		this.componentBindings = componentBindings;
		this.type = type;
	}
	
	public void postConstruction() 
	{
		for (int i=0; i<componentBindings.length; i++)
		{
			DataType componentType = ((RecordType)type).getComponentType(i);
			componentBindings[ i ] = createBindingUnsynchronized( componentType );
		}		
	}
	
	@Override
	public Object create(Object... value) {		
		return null;
	}

	@Override
	public Object getComponent(Object obj, int index) throws UnsupportedOperationException {
    	throw new UnsupportedOperationException();
	}

	@Override
	public Object createPartial() {
		return null;
	}

	@Override
	public void setComponents(Object obj, Object... value) {
	}
	
	@Override
	public void setComponent(Object obj, int index, Object value)
			throws BindingException {
	}

	@Override
	public boolean isInstance(Object obj) {
		return false;
	}
	
    @Override
    public int deepCompare(Object o1, Object o2,
    		Set<IdentityPair<Object, Object>> compareHistory)
    		throws BindingException {
		return 0;
	}
	
	
}

static class UnboundArray extends ArrayBinding {
	
	public UnboundArray(ArrayType type, Binding componentBinding) {
		super(componentBinding);
		this.type = type;
	}
	
	@Override
	public Object create() {
		return null;
	}
	
	@Override
	public Object create(int length, Iterator<Object> values) {
		return null;
	}
	
	@Override
	public Object create(Object...array) throws BindingException {
		return null;
	}

	@Override
	public Object get(Object array, int index) throws UnsupportedOperationException {
    	throw new UnsupportedOperationException();
	}
	
	@Override
	public void getAll(Object array, Object[] result) throws BindingException {
    	throw new UnsupportedOperationException();		
	}
	
	@Override
	public void set(Object array, int index, Object value)
			throws BindingException {
    	throw new UnsupportedOperationException();		
	}

	@Override
	public int size(Object array) throws UnsupportedOperationException {
    	throw new UnsupportedOperationException();
	}

	@Override
	public boolean isInstance(Object obj) {
		return false;
	}	

    @Override
    public int deepCompare(Object o1, Object o2,
    		Set<IdentityPair<Object, Object>> compareHistory)
    		throws BindingException {
		return 0;
	}
    
    @Override
    public void add(Object array, int index, Object element)
    		throws BindingException, IndexOutOfBoundsException {
    	throw new UnsupportedOperationException();
    }
    
    @Override
    public void remove(Object array, int index, int count) throws BindingException {
    	throw new UnsupportedOperationException();
    }
	
}

class UnboundUnion extends UnionBinding {
	
	public UnboundUnion(UnionType type, Binding[] componentBindings)
	{
		this.componentBindings = componentBindings;
		this.type = type;
	}

	public void postConstruction() 
	{
		for (int i=0; i<componentBindings.length; i++)
		{
			DataType componentType = ((UnionType)type).components[i].type;
			componentBindings[ i ] = createBindingUnsynchronized( componentType );
		}		
	}

	@Override
	public Object create(int tag, Object value) {
		return null;
	}
	
	@Override
	public void setValue(Object union, int tag, Object value)
			throws BindingException {
	}

	@Override
	public int getTag(Object obj) throws UnsupportedOperationException {
    	throw new UnsupportedOperationException();
	}

	@Override
	public Object getValue(Object obj) throws UnsupportedOperationException {
    	throw new UnsupportedOperationException();
	}

	@Override
	public boolean isInstance(Object obj) {
		return false;
	}
	
    @Override
    public int deepCompare(Object o1, Object o2,
    		Set<IdentityPair<Object, Object>> compareHistory)
    		throws BindingException {

		return 0;
	}

	
}

static class UnboundVariant extends VariantBinding {

	public UnboundVariant() {
		super();
	}

	public Object createFrom(Object value, Binding binding) throws BindingException {
		return null;
	}
	
	@Override
	public Object create(Object value, Binding binding) throws BindingException {
		return null;
	}

	@Override
	public Binding getBinding(Object variant) throws BindingException {
		return null;
	}

	@Override
	public DataType getType(Object variant) throws BindingException {
    	throw new UnsupportedOperationException();
	}

	@Override
	public Object getValue(Object variant, Binding binding)
			throws BindingException {
    	throw new UnsupportedOperationException();
	}
	
	@Override
	public Object getValue(Object variant)
	throws BindingException {
    	throw new UnsupportedOperationException();
	}

	@Override
	public void setValue(Object variant, Object value, Binding binding)
			throws BindingException {
	}

	@Override
	public void assertInstaceIsValid(Object obj, Set<Object> validInstances) throws BindingException {
	}

	@Override
	public boolean isInstance(Object obj) {
		return false;
	}
	
    @Override
    public int deepCompare(Object o1, Object o2,
    		Set<IdentityPair<Object, Object>> compareHistory)
    		throws BindingException {

		return 0;
	}

}

static class UnboundMap extends MapBinding {
	
	public UnboundMap(MapType type, Binding keyBinding, Binding valueBinding) {
		super(type, keyBinding, valueBinding);		
	}

	@Override
	public void clear(Object map) {
	}

	@Override
	public boolean containsKey(Object map, Object key) {
		return false;
	}

	@Override
	public boolean containsValue(Object map, Object value) {
		return false;
	}

	@Override
	public Object create() {
		return null;
	}

	@Override
	public Object create(Map initialMap) {
		return null;
	}

	@Override
	public Object create(Object[] keys, Object[] values) {
		return null;
	}
	
	@Override
	public Object create(List<Object> keys, List<Object> values)
			throws BindingException {
		return null;
	}

	@Override
	public Object get(Object map, Object key) throws UnsupportedOperationException {
    	throw new UnsupportedOperationException();
	}

	@Override
	public void getAll(Object mapFrom, Map to) throws UnsupportedOperationException {
    	throw new UnsupportedOperationException();
	}

	@Override
	public Object[] getKeys(Object map) throws UnsupportedOperationException {
    	throw new UnsupportedOperationException();
	}
	
	@Override
	public void getKeys(Object map, Set<Object> keys) throws BindingException {
    	throw new UnsupportedOperationException();
	}

	@Override
	public void getAll(Object mapFrom, Object[] keys, Object[] values) throws UnsupportedOperationException {		
    	throw new UnsupportedOperationException();
	}
	
	@Override
	public Object[] getValues(Object map) throws UnsupportedOperationException {
    	throw new UnsupportedOperationException();
	}

	@Override
	public void put(Object map, Object key, Object value) {
	}

	@Override
	public void putAll(Object mapTo, Map from) {
	}
	
	@Override
	public Object remove(Object map, Object key) throws UnsupportedOperationException {
    	throw new UnsupportedOperationException();
	}

	@Override
	public int size(Object map) throws UnsupportedOperationException {
    	throw new UnsupportedOperationException();
	}

	@Override
	public boolean isInstance(Object obj) {
		return false;
	}

    @Override
    public int deepCompare(Object o1, Object o2,
    		Set<IdentityPair<Object, Object>> compareHistory)
    		throws BindingException {

		return 0;
	}

	@Override
	public Object getCeilingKey(Object map, Object key) {
		return null;
	}

	@Override
	public Object getFirstKey(Object map) {
		return null;
	}

	@Override
	public Object getFloorKey(Object map, Object key) {
		return null;
	}

	@Override
	public Object getHigherKey(Object map, Object key) {
		return null;
	}

	@Override
	public Object getLastKey(Object map) {
		return null;
	}

	@Override
	public Object getLowerKey(Object map, Object key) {
		return null;
	}
	
}

/**
 * No value can be assigned to unbound optional binding. 
 * The object is always null.
 * 
 */
public static class UnboundOptionalBinding extends OptionalBinding {

	public UnboundOptionalBinding(OptionalType componentType) {
		super(componentType, null);
	}
	
	@Override
	public Object createNoValue() {
		return null;
	}

	@Override
	public Object createValue(Object value) throws BindingException {
		return null;
	}

	@Override
	public Object getValue(Object arg) throws BindingException {
		if (arg==null) return null;
		throw new BindingException("Not supported");
	}

	@Override
	public boolean hasValue(Object arg) {		
		return false;
	}

	@Override
	public boolean isInstance(Object obj) {
		return obj==null;
	}

	@Override
	public void setNoValue(Object optional) throws BindingException {		
	}

	@Override
	public void setValue(Object optional, Object componentValue)
			throws BindingException {
	}
	
}

}