/*******************************************************************************
 * Copyright (c) 2010- Association for Decentralized Information Management in
 * Industry THTH ry.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors:
 *    VTT Technical Research Centre of Finland - initial API and implementation
 *******************************************************************************/
package org.simantics.databoard.history;

import java.io.IOException;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.util.Set;

import org.simantics.databoard.Bindings;
import org.simantics.databoard.binding.Binding;
import org.simantics.databoard.binding.IntegerBinding;
import org.simantics.databoard.binding.LongBinding;
import org.simantics.databoard.binding.StringBinding;
import org.simantics.databoard.binding.VariantBinding;
import org.simantics.databoard.binding.error.BindingException;
import org.simantics.databoard.binding.java.IntegerBindingDefault;
import org.simantics.databoard.binding.java.LongBindingDefault;
import org.simantics.databoard.binding.java.StringBindingDefault;
import org.simantics.databoard.binding.mutable.Variant;
import org.simantics.databoard.binding.mutable.VariantContainerBinding;
import org.simantics.databoard.serialization.SerializationException;
import org.simantics.databoard.serialization.Serializer;
import org.simantics.databoard.type.DataType;
import org.simantics.databoard.type.IntegerType;
import org.simantics.databoard.type.LongType;
import org.simantics.databoard.type.StringType;
import org.simantics.databoard.util.Base64;
import org.simantics.databoard.util.URIUtil;
import org.simantics.databoard.util.binary.BinaryMemory;
import org.simantics.databoard.util.binary.BinaryReadable;
import org.simantics.databoard.util.binary.RuntimeSerializerConstructionException;

/**
 * This class binding Variant to a filename/URL compatible String.
 * The value is human-readable for strings, integers and longs. 
 * For datatypes the value is Base64 encoded binary.
 * 
 *  Filenames have the following encoding:
 *    S<string>       String types, if string doesn't have the following 
 *                    control characters " : < > | ? * # \ / % [0..31] [128..] are escaped with %<hex><hex>
 *    I<integer>      Integer types
 *    L<long>         Long types
 *    B<base64>       All other cases. The value is binary encoded and presented as single line Base64 string.
 *                    Base64 encoding has url and filename safe options enabled. 
 *
 * See StringVariantBindingExample
 * @author Toni Kalajainen <toni.kalajainen@vtt.fi>
 */
public class StringVariantBinding extends VariantBinding {

	static final Serializer DATATYPE_SERIALIZER = Bindings.getBindingUnchecked(DataType.class).serializer();  
	
    public static final char[] HEX_VALUES = {'0', '1', '2', '3' ,'4' ,'5', '6', '7', '8', '9' ,'A', 'B', 'C', 'D', 'E', 'F'};
	
	public static final StringVariantBinding INSTANCE = new StringVariantBinding();
	
	@Override
	public Object create(Object value, Binding binding) throws BindingException {
		if (binding instanceof StringBinding && binding.type().equals(StringType.INSTANCE)) {
			StringBinding sb = (StringBinding) binding;
			return "S"+URIUtil.encodeURI( sb.getValue(value) );
		}
		
		if (binding instanceof IntegerBinding && binding.type().equals(IntegerType.INSTANCE)) {
			IntegerBinding ib = (IntegerBinding) binding;
			return "I"+ib.getValue_(value);
		}
		
		if (binding instanceof LongBinding && binding.type().equals(LongType.INSTANCE)) {
			LongBinding lb = (LongBinding) binding;
			return "L"+lb.getValue_(value);
		}
		
		try {
			Variant v = new Variant(binding, value);
			byte[] data = VariantContainerBinding.INSTANCE.serializer().serialize(v);
			return "B"+Base64.encodeBytes(data, Base64.URL_SAFE);
		} catch (RuntimeSerializerConstructionException e) {
			throw new BindingException(e);
		} catch (SerializationException e) {
			throw new BindingException(e);
		} catch (IOException e) {
			throw new BindingException(e);
		}
	}

	@Override
	public Binding getBinding(Object variant) throws BindingException {
		String str = (String) variant;
		if (str.startsWith("S")) {
			return StringBindingDefault.INSTANCE;
		}
		if (str.startsWith("I")) {
			return IntegerBindingDefault.INSTANCE;
		}
		if (str.startsWith("L")) {
			return LongBindingDefault.INSTANCE;
		}
		if (str.startsWith("B")) {
			try {
				byte[] 	data = Base64.decode(str.substring(1), Base64.URL_SAFE);
				BinaryReadable readable = new BinaryMemory( ByteBuffer.wrap(data) ) ;
				readable.order(ByteOrder.LITTLE_ENDIAN);
				DataType type = (DataType) DATATYPE_SERIALIZER.deserialize(readable);
				return Bindings.getBinding(type);
			} catch (IOException e) {
				throw new BindingException(e);
			} catch (SerializationException e) {
				throw new BindingException(e);
			}			
		}
		
		throw new BindingException("Invalid string");
	}

	@Override
	public DataType getType(Object variant) throws BindingException {
		String str = (String) variant;
		if (str.startsWith("S")) {
			return StringBindingDefault.INSTANCE.type();
		}
		if (str.startsWith("I")) {
			return IntegerBindingDefault.INSTANCE.type();
		}
		if (str.startsWith("L")) {
			return LongBindingDefault.INSTANCE.type();
		}
		if (str.startsWith("B")) {
			try {
				byte[] 	data = Base64.decode(str.substring(1), Base64.URL_SAFE);
				BinaryReadable readable = new BinaryMemory( ByteBuffer.wrap(data) ) ;
				readable.order(ByteOrder.LITTLE_ENDIAN);
				return (DataType) DATATYPE_SERIALIZER.deserialize(readable);
			} catch (IOException e) {
				throw new BindingException(e);
			} catch (SerializationException e) {
				throw new BindingException(e);
			}			
		}
		
		throw new BindingException("Invalid string");
	}

	@Override
	public Object getValue(Object obj, Binding binding)
			throws BindingException {
		if (obj==null) throw new BindingException("null value is not Variant");
		if (obj instanceof String == false) throw new BindingException("wrong class, String expected");
		
		String str = (String) obj;
		
		if (str.startsWith("S")) {
			if (binding instanceof StringBinding == false) throw new BindingException("StringBinding expected, got "+binding.getClass().getSimpleName());
			String value = URIUtil.decodeURI( str.substring(1) );
			StringBinding sb = (StringBinding) binding;
			return sb.create(value);
		}
		
		if (str.startsWith("I")) {
			if (binding instanceof IntegerBinding == false) throw new BindingException("IntegerBinding expected, got "+binding.getClass().getSimpleName());
			try {
				Integer value = Integer.valueOf( str.substring(1) );
				IntegerBinding ib = (IntegerBinding) binding;
				return ib.create(value);				
			} catch (NumberFormatException nfe) {
				throw new BindingException(nfe);
			}
		}

		if (str.startsWith("L")) {
			if (binding instanceof LongBinding == false) throw new BindingException("LongBinding expected, got "+binding.getClass().getSimpleName());
			try {
				Long value = Long.valueOf( str.substring(1) );
				LongBinding lb = (LongBinding) binding;
				return lb.create(value);				
			} catch (NumberFormatException nfe) {
				throw new BindingException(nfe);
			}
		}
		
		if (str.startsWith("B")) {
			try {
				byte[] data = Base64.decode(str.substring(1), Base64.URL_SAFE);
				BinaryReadable readable = new BinaryMemory( ByteBuffer.wrap(data) ) ;
				readable.order(ByteOrder.LITTLE_ENDIAN);
				DataType type = (DataType) DATATYPE_SERIALIZER.deserialize(readable);
				if (!type.equals(binding.type()))
					throw new BindingException("Binding for "+type.toSingleLineString()+" expected, but got "+binding.type());
				return binding.serializer().deserialize(readable);
			} catch (IOException ioe) {
				throw new BindingException(ioe);
			} catch (RuntimeSerializerConstructionException e) {
				throw new BindingException(e);
			} catch (SerializationException e) {
				throw new BindingException(e);
			}
		}
		
		throw new BindingException("Invalid value");		
	}

	@Override
	public Object getValue(Object obj) throws BindingException {
		if (obj==null) throw new BindingException("null value is not Variant");
		if (obj instanceof String == false) throw new BindingException("wrong class, String expected");
		
		String str = (String) obj;
		
		if (str.startsWith("S")) {
			String value = URIUtil.decodeURI( str.substring(1) );			
			return StringBindingDefault.INSTANCE.create(value);
		}
		
		if (str.startsWith("I")) {
			try {
				Integer value = Integer.valueOf( str.substring(1) );
				return IntegerBindingDefault.INSTANCE.create(value);				
			} catch (NumberFormatException nfe) {
				throw new BindingException(nfe);
			}
		}

		if (str.startsWith("L")) {
			try {
				Long value = Long.valueOf( str.substring(1) );
				return LongBindingDefault.INSTANCE.create(value);				
			} catch (NumberFormatException nfe) {
				throw new BindingException(nfe);
			}
		}
		
		if (str.startsWith("B")) {
			try {
				byte[] data = Base64.decode(str.substring(1), Base64.URL_SAFE);
				BinaryReadable readable = new BinaryMemory( ByteBuffer.wrap(data) );
				Variant variant = (Variant) VariantContainerBinding.INSTANCE.serializer().deserialize(readable);
				return variant.getValue();
			} catch (IOException ioe) {
				throw new BindingException(ioe);
			} catch (RuntimeSerializerConstructionException e) {
				throw new BindingException(e);
			} catch (SerializationException e) {
				throw new BindingException(e);
			}
		}
		
		throw new BindingException("Invalid value");		
	}

	@Override
	public void setValue(Object variant, Object value, Binding binding)
			throws BindingException {
		throw new BindingException("Cannot set valut of an immutable String");
	}

	@Override
	public void assertInstaceIsValid(Object obj, Set<Object> validInstances)
			throws BindingException {
		if (obj==null) throw new BindingException("null value is not Variant");
		if (obj instanceof String == false) throw new BindingException("wrong class, String expected");
		Object value = getValue(obj);
		Binding binding = getBinding(obj);
		binding.assertInstaceIsValid(value);
	}

	@Override
	public boolean isInstance(Object obj) {		
		return obj instanceof String;
	}

}

