/*******************************************************************************
 *  Copyright (c) 2010 Association for Decentralized Information Management in
 *  Industry THTH ry.
 *  All rights reserved. This program and the accompanying materials
 *  are made available under the terms of the Eclipse Public License v1.0
 *  which accompanies this distribution, and is available at
 *  http://www.eclipse.org/legal/epl-v10.html
 *
 *  Contributors:
 *      VTT Technical Research Centre of Finland - initial API and implementation
 *******************************************************************************/
package org.simantics.databoard.reflection;

import java.lang.annotation.Annotation;
import java.lang.reflect.Field;
import java.lang.reflect.Modifier;
import java.lang.reflect.ParameterizedType;
import java.lang.reflect.Type;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;
import java.util.TreeSet;

import org.simantics.databoard.annotations.ComponentType;
import org.simantics.databoard.annotations.KeyType;
import org.simantics.databoard.annotations.Length;
import org.simantics.databoard.annotations.MIMEType;
import org.simantics.databoard.annotations.Optional;
import org.simantics.databoard.annotations.Pattern;
import org.simantics.databoard.annotations.Range;
import org.simantics.databoard.annotations.Referable;
import org.simantics.databoard.annotations.Union;
import org.simantics.databoard.annotations.Unit;
import org.simantics.databoard.annotations.ValueType;
import org.simantics.databoard.binding.ArrayBinding;
import org.simantics.databoard.binding.Binding;
import org.simantics.databoard.binding.BooleanBinding;
import org.simantics.databoard.binding.ByteBinding;
import org.simantics.databoard.binding.DoubleBinding;
import org.simantics.databoard.binding.FloatBinding;
import org.simantics.databoard.binding.IntegerBinding;
import org.simantics.databoard.binding.LongBinding;
import org.simantics.databoard.binding.MapBinding;
import org.simantics.databoard.binding.OptionalBinding;
import org.simantics.databoard.binding.StringBinding;
import org.simantics.databoard.binding.error.BindingConstructionException;
import org.simantics.databoard.binding.java.ArrayListBinding;
import org.simantics.databoard.binding.java.BooleanBindingDefault;
import org.simantics.databoard.binding.java.ByteBindingDefault;
import org.simantics.databoard.binding.java.DoubleBindingDefault;
import org.simantics.databoard.binding.java.FloatBindingDefault;
import org.simantics.databoard.binding.java.HashMapBinding;
import org.simantics.databoard.binding.java.HashSetBinding;
import org.simantics.databoard.binding.java.IntegerBindingDefault;
import org.simantics.databoard.binding.java.LinkedListBinding;
import org.simantics.databoard.binding.java.LongBindingDefault;
import org.simantics.databoard.binding.java.OptionalBindingDefault;
import org.simantics.databoard.binding.java.StringBindingDefault;
import org.simantics.databoard.binding.java.TreeMapBinding;
import org.simantics.databoard.binding.java.TreeSetBinding;
import org.simantics.databoard.binding.mutable.MutableBoolean;
import org.simantics.databoard.binding.mutable.MutableBooleanBinding;
import org.simantics.databoard.binding.mutable.MutableByte;
import org.simantics.databoard.binding.mutable.MutableByteBinding;
import org.simantics.databoard.binding.mutable.MutableDouble;
import org.simantics.databoard.binding.mutable.MutableDoubleBinding;
import org.simantics.databoard.binding.mutable.MutableFloat;
import org.simantics.databoard.binding.mutable.MutableFloatBinding;
import org.simantics.databoard.binding.mutable.MutableInteger;
import org.simantics.databoard.binding.mutable.MutableIntegerBinding;
import org.simantics.databoard.binding.mutable.MutableLong;
import org.simantics.databoard.binding.mutable.MutableLongBinding;
import org.simantics.databoard.binding.mutable.MutableString;
import org.simantics.databoard.binding.mutable.MutableStringBinding;
import org.simantics.databoard.binding.mutable.Variant;
import org.simantics.databoard.binding.mutable.VariantContainerBinding;
import org.simantics.databoard.bindingscheme.DataTypeConstructionException;
import org.simantics.databoard.type.ArrayType;
import org.simantics.databoard.type.BooleanType;
import org.simantics.databoard.type.ByteType;
import org.simantics.databoard.type.Component;
import org.simantics.databoard.type.DataType;
import org.simantics.databoard.type.DoubleType;
import org.simantics.databoard.type.FloatType;
import org.simantics.databoard.type.IntegerType;
import org.simantics.databoard.type.LongType;
import org.simantics.databoard.type.MapType;
import org.simantics.databoard.type.OptionalType;
import org.simantics.databoard.type.RecordType;
import org.simantics.databoard.type.StringType;
import org.simantics.databoard.type.UnionType;
import org.simantics.databoard.type.VariantType;
import org.simantics.databoard.util.RangeException;

/**
 * Reflection Binding creates bindings and types of Java Classes.
 * 
 * o See org.simantics.databoard.annotations for type-class annotations.
 * o List, ArrayList, primitive[] are bound to Array Type 
 * o Map, HashMap and TreeMap are bound to Map(?, ?)
 * o Set, TreeSet and HashSet are bound to Map(T, {})
 * o Void.class is bound to an empty record {}.
 * o Object.class is bound to an immutable variant.
 * o Exception StackTraces are omited. 
 * o Classes are bound to a Record Type. There are three types classes supported: 
 *  1) record-like class 
 *     - All fields are public
 *     - No-argument public constructor
 *     
 *  2) immutable-like class
 *     - All fields are assigneed in the constructor
 * 
 *  3) bean-like class
 *     - All fields are set with getter/setter
 *      
 *  
 * 
 * @author Toni Kalajainen <toni.kalajainen@vtt.fi>
 */
public class ReflectionBinding {

	HashMap<ClassDataType, DataType> dataTypeCache = 
		new HashMap<ClassDataType, DataType>();
	
	HashMap<ClassBinding, Binding> bindingCache = 
		new HashMap<ClassBinding, Binding>();

	public ReflectionBinding() {
		dataTypeCache.put(new ClassDataType(Integer.class), IntegerType.INSTANCE);
		dataTypeCache.put(new ClassDataType(int.class), IntegerType.INSTANCE);
		dataTypeCache.put(new ClassDataType(Byte.class), ByteType.INSTANCE);
		dataTypeCache.put(new ClassDataType(byte.class), ByteType.INSTANCE);
		dataTypeCache.put(new ClassDataType(Float.class), FloatType.INSTANCE);
		dataTypeCache.put(new ClassDataType(float.class), FloatType.INSTANCE);
		dataTypeCache.put(new ClassDataType(Double.class), DoubleType.INSTANCE);
		dataTypeCache.put(new ClassDataType(double.class), DoubleType.INSTANCE);
		dataTypeCache.put(new ClassDataType(Long.class), LongType.INSTANCE);
		dataTypeCache.put(new ClassDataType(long.class), LongType.INSTANCE);
		dataTypeCache.put(new ClassDataType(String.class), StringType.INSTANCE);
		dataTypeCache.put(new ClassDataType(Boolean.class), BooleanType.INSTANCE);
		dataTypeCache.put(new ClassDataType(boolean.class), BooleanType.INSTANCE);	
		dataTypeCache.put(new ClassDataType(Object.class), VariantType.INSTANCE);	
		
		// void.class to VoidBinding
		{
			Class<?> clazz = void.class;
			VoidBinding binding = new VoidBinding();
			DataType type = binding.type();
			ClassDataType classType = new ClassDataType(clazz);
			ClassBinding classBinding = new ClassBinding(type, classType);
			dataTypeCache.put(classType, type);
			bindingCache.put(classBinding, binding);
		}
		
		// Variant.class to VariantContainerBinding
		{
			DataType type = VariantType.INSTANCE;
			Class<?> clazz = Variant.class;
			Binding binding = VariantContainerBinding.INSTANCE;
			ClassDataType classType = new ClassDataType(clazz);
			ClassBinding classBinding = new ClassBinding(type, classType);
			dataTypeCache.put(classType, type);	
			bindingCache.put(classBinding, binding);
		}
		
		
	}
	
	
	/**
	 * Read representation from a class. DataType details and parameters
	 * are read as annotations placed in the class.  
	 * (See org.simantics.databoard.annotations)  
	 * <p>
	 * As an exception, in the subclasses of {@link Throwable}, the fields of
	 * Throwable are omited.
	 * 
	 * @param clazz
	 * @return data type
	 */
	public synchronized DataType getDataType(Class<?> clazz) 
	throws DataTypeConstructionException
	{
		return unsynchronizedGetDataType(new ClassDataType(clazz));
	}
    
	/**
	 * Read binding and type from a class. DataType details and parameters
	 * are read as annotations placed in the class.  
	 * (See org.simantics.databoard.annotations)
	 * <p>
	 * As an exception, in the subclasses of {@link Throwable}, the fields of
	 * Throwable are omited.
	 * 
	 * @param clazz
	 * @return binding
	 * @throws BindingConstructionException
	 */
    public synchronized Binding getBinding(Class<?> clazz) 
    throws BindingConstructionException
    {
    	ClassDataType dt = new ClassDataType(clazz);
		DataType dataType;
		try {
			dataType = unsynchronizedGetDataType(dt);
		} catch (DataTypeConstructionException e) {
			throw new BindingConstructionException(e);
		}
		ClassBinding bindingRequest = new ClassBinding(dataType, dt);
		return unsynchronizedGetBinding(bindingRequest, bindingCache);
    }

    DataType unsynchronizedGetDataType(ClassDataType clazz) 
    throws DataTypeConstructionException
    {
    	if (dataTypeCache.containsKey(clazz)) 
    		return dataTypeCache.get(clazz);
    	
    	Optional optional = clazz.getAnnotation(Optional.class); 
    	if(optional != null)
    	{    		
    		ClassDataType componentClassRequest = clazz.dropAnnotation(optional); 
    		DataType componentType = unsynchronizedGetDataType(componentClassRequest);
    		OptionalType type = new OptionalType(componentType);
            dataTypeCache.put(clazz, type);                        
    		return type;
    	}
        
    	// Primitive
    	{
			try {  		
				Range range = clazz.getAnnotation(Range.class);
				Unit unit = clazz.getAnnotation(Unit.class);
				
				if (clazz.clazz == Integer.class || clazz.clazz == int.class || clazz.clazz == MutableInteger.class) {
					if (range==null && unit==null) return IntegerType.INSTANCE;
					IntegerType result = new IntegerType();
					result.setRange( range==null ? null : org.simantics.databoard.util.Range.valueOf(range.value()) );
					result.unit = unit==null ? null : unit.value();
		            dataTypeCache.put(clazz, result);                        
					return result;
				}
					
				if (clazz.clazz == Byte.class || clazz.clazz == byte.class || clazz.clazz == MutableByte.class) {
					if (range==null && unit==null) return ByteType.INSTANCE;
					ByteType result = new ByteType();
					result.setRange( range==null ? null : org.simantics.databoard.util.Range.valueOf(range.value()) );
					result.unit = unit==null ? null : unit.value();
		            dataTypeCache.put(clazz, result);                        
					return result;
				}			
				
				if (clazz.clazz == Long.class || clazz.clazz == long.class || clazz.clazz == MutableLong.class) {
					if (range==null && unit==null) return LongType.INSTANCE;
					LongType result = new LongType();
					result.setRange( range==null ? null : org.simantics.databoard.util.Range.valueOf(range.value()) );
					result.unit = unit==null ? null : unit.value();
		            dataTypeCache.put(clazz, result);                        
					return result;
				}
				
				if (clazz.clazz == Float.class || clazz.clazz == float.class || clazz.clazz == MutableFloat.class) {
					if (range==null && unit==null) return FloatType.INSTANCE;
					FloatType result = new FloatType();
					result.setRange( range==null ? null : org.simantics.databoard.util.Range.valueOf(range.value()) );
					result.unit = unit==null ? null : unit.value();
		            dataTypeCache.put(clazz, result);                        
					return result;
				}
				
				if (clazz.clazz == Double.class || clazz.clazz == double.class || clazz.clazz == MutableDouble.class) {
					if (range==null && unit==null) return DoubleType.INSTANCE;
					DoubleType result = new DoubleType();
					result.setRange( range==null ? null : org.simantics.databoard.util.Range.valueOf(range.value()) );
					result.unit = unit==null ? null : unit.value();
		            dataTypeCache.put(clazz, result);                        
					return result;
				}
	
				if (clazz.clazz == Boolean.class || clazz.clazz == boolean.class || clazz.clazz == MutableBoolean.class) {
					return BooleanType.INSTANCE;
				}

				if (clazz.clazz == String.class || clazz.clazz == MutableString.class) {
					Length length = clazz.getAnnotation(Length.class);
					MIMEType mimeType = clazz.getAnnotation(MIMEType.class);
					Pattern pattern = clazz.getAnnotation(Pattern.class);
					if (length==null && mimeType==null && pattern==null)
						return StringType.INSTANCE;
					StringType str = new StringType();
					str.setLength( length==null ? null : org.simantics.databoard.util.Range.valueOf( length.value()[0] ) );
					str.mimeType = mimeType==null ? null : mimeType.value();
					str.pattern = pattern==null ? null : pattern.value();					
		            dataTypeCache.put(clazz, str);                        
					return str;
				}
				
			} catch (RangeException e) {
				throw new DataTypeConstructionException(e);
			}
		}
		
        // Array or List type
        if(clazz.clazz.isArray() || clazz.clazz.equals(ArrayList.class) || clazz.clazz.equals(LinkedList.class) || List.class.isAssignableFrom(clazz.clazz) ) {
        	try {
        		org.simantics.databoard.util.Range[] lengths = null;
        		Annotation[] annotations = clazz.annotations.clone();
        		
        		Class<?> componentType = clazz.clazz.getComponentType();
        		
	            if (annotations!=null) {
	            	
	            	// Create a new annotation list, basicly removes Length element if one exists
	            	List<Annotation> newList = new ArrayList<Annotation>(annotations.length);
	            	
		            for (Annotation a : annotations) {
		            	if (a instanceof Length) {
		            		String[] strs = ((Length)a).value();
		            		lengths = new org.simantics.databoard.util.Range[strs.length];
		            	    for (int i=0; i<strs.length; i++)
		            	    	lengths[i] = org.simantics.databoard.util.Range.valueOf(strs[i]);
		            	} else if (a instanceof ComponentType) {
		            		componentType = ((ComponentType)a).value();
		            	} else {
		            		newList.add(a);
		            	}
		            }
		            
		            annotations = newList.toArray( new Annotation[newList.size()] );
	            }
            
	            if (componentType==null) {
	            	throw new DataTypeConstructionException("Cannot determine component type");
	            }
	            
	            ArrayType type = new ArrayType();
	            type.componentType = unsynchronizedGetDataType( new ClassDataType( componentType, annotations ) );
	
	            // Multi-dimensional Array. Apply range to sub-array-types
	            ArrayType t = type;
	            if (lengths!=null)
	                for (int i=0; i<lengths.length; i++) {
	                	t.setLength( lengths[i] );
	                    if (i<lengths.length-1)
	                        t = (ArrayType) t.componentType;
	                }
	            dataTypeCache.put(clazz, type);                        
	            return type;
        	} catch (RangeException e) {
        		throw new DataTypeConstructionException(e);
        	}
        }
        
        // Set Type
        if ( Set.class.isAssignableFrom(clazz.clazz) || clazz.clazz.equals(TreeSet.class) || clazz.clazz.equals(HashSet.class) ) {
    		Annotation[] annotations = clazz.annotations.clone();    		
    		Class<?> keyClass = null;
    		
            if (annotations!=null) {
            	
            	// Create a new annotation list, basicly removes Length element if one exists
            	List<Annotation> newList = new ArrayList<Annotation>(annotations.length);
            	
	            for (Annotation a : annotations) {
	            	if (a instanceof KeyType) {
	            		keyClass = ((KeyType)a).value();
	            	} else {
	            		newList.add(a);
	            	}
	            }
	            
	            annotations = newList.toArray( new Annotation[newList.size()] );
            }
        
            if (keyClass==null) {
            	throw new DataTypeConstructionException("Cannot determine key class");
            }
            
            DataType keyType = unsynchronizedGetDataType( new ClassDataType( keyClass, annotations ) );
            DataType valueType = unsynchronizedGetDataType( new ClassDataType( void.class ) );
            MapType type = new MapType(keyType, valueType);
            dataTypeCache.put(clazz, type);                        
            return type;
        }
        
        // Map Type
        if ( Map.class.isAssignableFrom(clazz.clazz) || clazz.clazz.equals(TreeMap.class) || clazz.clazz.equals(HashMap.class) ) {
    		Annotation[] annotations = clazz.annotations.clone();    		
    		Class<?> keyClass = null;
    		Class<?> valueClass = null;
    		
            if (annotations!=null) {
            	
            	// Create a new annotation list, basicly removes Length element if one exists
            	List<Annotation> newList = new ArrayList<Annotation>(annotations.length);
            	
	            for (Annotation a : annotations) {
	            	if (a instanceof ValueType) {
	            		valueClass = ((ValueType)a).value();
	            	} else if (a instanceof KeyType) {
	            		keyClass = ((KeyType)a).value();
	            	} else {
	            		newList.add(a);
	            	}
	            }
	            
	            annotations = newList.toArray( new Annotation[newList.size()] );
            }
        
            if (keyClass==null||valueClass==null) {
            	throw new DataTypeConstructionException("Cannot determine key and value class");
            }
            
            DataType keyType = unsynchronizedGetDataType( new ClassDataType( keyClass, annotations ) );
            DataType valueType = unsynchronizedGetDataType( new ClassDataType( valueClass, annotations ) );
            MapType type = new MapType(keyType, valueType);
            dataTypeCache.put(clazz, type);                        
            return type;
        }
        
        if (clazz.clazz.isEnum()) {
            Enum<?>[] enums = (Enum[]) clazz.clazz.getEnumConstants();
            UnionType type = new UnionType();
//            type.name = clazz.getSimpleName();
            type.components = new Component[enums.length];
            for (int i=0; i<enums.length; i++) {
                String name = enums[i].name();
                type.components[i] = new Component(
                		name,
                		new RecordType(false)
                		);                
            }            
            dataTypeCache.put(clazz, type);                        
            return type;
        }
        
        // Union type
        Union union = clazz.getAnnotation(Union.class);
        if(union != null) {
        	UnionType type = new UnionType();
            Class<?>[] cases = union.value();
            type.components = new Component[cases.length];
            for(int i=0;i<cases.length;++i)
                type.components[i] = new Component(
                		cases[i].getSimpleName(), null /* updated later*/);
            
            // Hash code is correct NOW
            dataTypeCache.put(clazz, type);
            for(int i=0;i<cases.length;++i) {
            	Component component = type.components[i];
                component.type = unsynchronizedGetDataType(new ClassDataType( cases[i] ));
            }
            return type;
        }
        
        // Record type
        {
        	RecordType type = new RecordType();
//        	type.name = clazz.getSimpleName();
            Field[] fields = getFields(clazz.clazz);
            Component[] components = new Component[fields.length];
            for(int i=0;i<fields.length;++i) {
            	Field field = fields[i];
                components[i] = new Component(field.getName(), null /* updated later */);
            }
            type.referable = clazz.getAnnotation(Referable.class) != null;
            type.setComponents( components );
            // hash is correct NOW
        	dataTypeCache.put(clazz, type);
            for(int i=0;i<type.count();++i) {
            	Component c = type.getComponent(i);
            	Field field = fields[i];
            	Class<?> fieldClass = field.getType(); 
            	Annotation[] annotations = getFieldAnnotations(field);            	            
            	ClassDataType classDef = new ClassDataType(fieldClass, annotations);            	
            	c.type = unsynchronizedGetDataType( classDef );
            }
            
            return type;
        }
    }
    
    
    @SuppressWarnings("unchecked")
    Binding unsynchronizedGetBinding(ClassBinding bindingRequest, Map<ClassBinding, Binding> map) 
    throws BindingConstructionException {
    	DataType dataType = bindingRequest.type; 
    	ClassDataType clazz = bindingRequest.clazz;
    	
		try {
	    	if (dataType == null) throw new IllegalArgumentException("null arg");
	    	
	    	{
	    		Binding result = map.get(bindingRequest);
	    		if (result != null) return result;
	    		result = bindingCache.get(clazz);
	    		if (result != null) return result;
	    	}
	   		
	    	if (dataType instanceof IntegerType) {
	    		IntegerBinding result;
	    		if (clazz.clazz == MutableInteger.class) 
	    			result = new MutableIntegerBinding((IntegerType)dataType);
	    		else if (dataType.equals(IntegerBindingDefault.INSTANCE.type()))
	    			result = IntegerBindingDefault.INSTANCE;
	    		else 
	    			result = new IntegerBindingDefault((IntegerType)dataType);     		
	        	map.put(bindingRequest, result);
	        	return result;
	    	} else 
	    	if (dataType instanceof ByteType) {
	    		ByteBinding result;
	    		if (clazz.clazz == MutableByte.class) 
	    			result = new MutableByteBinding((ByteType)dataType);    		
	    		else if (dataType.equals( ByteBindingDefault.INSTANCE.type() ))
	    			result = ByteBindingDefault.INSTANCE;
	    		else 
	    			result = new ByteBindingDefault((ByteType)dataType);
	        	map.put(bindingRequest, result);
	        	return result;
	    	} else
	    	if (dataType instanceof DoubleType) {
	    		DoubleBinding result;
	    		if (clazz.clazz == MutableDouble.class) 
	    			result = new MutableDoubleBinding((DoubleType)dataType);    		
	    		else if (dataType.equals( DoubleBindingDefault.INSTANCE.type() ) )
	    			result = DoubleBindingDefault.INSTANCE;
	    		else
	    			result = new DoubleBindingDefault((DoubleType)dataType);
	        	map.put(bindingRequest, result);
	        	return result;
	    	} else
	    	if (dataType instanceof LongType) {
	    		LongBinding result;
	    		if (clazz.clazz == MutableLong.class) 
	    			result = new MutableLongBinding((LongType)dataType);    		
	    		else if (dataType.equals( LongBindingDefault.INSTANCE.type() ) )
	    			result = LongBindingDefault.INSTANCE;
	    		else 
	    			result = new LongBindingDefault((LongType)dataType);
	        	map.put(bindingRequest, result);
	        	return result;
	    	} else
	    	if (dataType instanceof FloatType) {
	    		FloatBinding result;
	    		if (clazz.clazz == MutableFloat.class) 
	    			result = new MutableFloatBinding((FloatType)dataType);    		
	    		else if (dataType.equals( FloatBindingDefault.INSTANCE.type() ) )
	    			result = FloatBindingDefault.INSTANCE;
	    		else
	    			result = new FloatBindingDefault((FloatType)dataType);
	        	map.put(bindingRequest, result);
	        	return result;
	    	} else
	    	if (dataType instanceof BooleanType) {
	    		BooleanBinding result;
	    		if (clazz.clazz == MutableBoolean.class) 
	    			result = new MutableBooleanBinding((BooleanType)dataType);    		
	    		else if (dataType.equals( BooleanBindingDefault.INSTANCE.type() ))
	    			result = BooleanBindingDefault.INSTANCE;
	    		else
	    			result = new BooleanBindingDefault((BooleanType)dataType);
	        	map.put(bindingRequest, result);
	        	return result;
	    	} else
	    	if (dataType instanceof StringType) {
	    		StringBinding result;
	    		if (clazz.clazz == MutableString.class) 
	    			result = new MutableStringBinding((StringType)dataType);    		
	    		else if (dataType.equals( StringBindingDefault.INSTANCE.type()) )
	    			result = StringBindingDefault.INSTANCE;
	    		else
	    			result = new StringBindingDefault((StringType)dataType);
	        	map.put(bindingRequest, result);
	        	return result;
	    	} else
	    	if (dataType instanceof ArrayType) {
	    		
	    		if ( bindingRequest.clazz.clazz.isArray() ) {
	
	        		ArrayBinding result;
	        		ArrayType type = (ArrayType) dataType;
	        		Class<?> componentClass = clazz.clazz.getComponentType();    
	        		ClassBinding componentBindingRequest = new ClassBinding(type.componentType, new ClassDataType(componentClass));
	        		Binding componentBinding = unsynchronizedGetBinding(componentBindingRequest, map);
	        		result = new ReflectionArrayBinding(bindingRequest.clazz.clazz, type, componentBinding);
	            	map.put(bindingRequest, result);        	
	            	return result;
	    			
	    		}
	    		
	    		{
	        		ArrayBinding result;
	        		ArrayType type = (ArrayType) dataType;
	        		Class<?> componentClass = clazz.clazz.getComponentType();  
	       			for (Annotation a : bindingRequest.clazz.annotations)
	       				if (a instanceof ComponentType)
	       					componentClass = ((ComponentType)a).value();
	        		
	        		ClassBinding componentBindingRequest = new ClassBinding(type.componentType, new ClassDataType(componentClass));
	        		Binding componentBinding = unsynchronizedGetBinding(componentBindingRequest, map);
	
	        		if (List.class.equals( bindingRequest.clazz.clazz ) || ArrayList.class.equals( bindingRequest.clazz.clazz )) {
	           			result = new ArrayListBinding(type, componentBinding);
	        		} else if (LinkedList.class.equals( bindingRequest.clazz.clazz )) {
	           			result = new LinkedListBinding(type, componentBinding);
	        		} else {
	        			throw new BindingConstructionException("I don't know how to bind "+bindingRequest.clazz.clazz);    			
	        		}
	        		
	            	map.put(bindingRequest, result);        	
	            	return result;    			
	    		}
	    		
	    	} else
	       	if (dataType instanceof MapType) {
	       		
	   			MapBinding result;
	   			MapType type = (MapType) dataType;
	    			
	   			Class<?> keyClass = null;
	   			Class<?> valueClass = null;
	   			for (Annotation a : bindingRequest.clazz.annotations) {
	   				if (a instanceof KeyType) keyClass = ((KeyType)a).value();
	   				if (a instanceof ValueType) valueClass = ((ValueType)a).value();
	   			}
	   			
	   			if (keyClass==null) throw new BindingConstructionException("Cannot determine key or value class");
	   			ClassBinding keyBindingRequest = new ClassBinding(type.getKeyType(), new ClassDataType(keyClass));
	   			Binding keyBinding = unsynchronizedGetBinding(keyBindingRequest, map);
	   			
	   			if (Set.class.equals( bindingRequest.clazz.clazz ) || TreeSet.class.equals( bindingRequest.clazz.clazz )) {
	       			result = new TreeSetBinding(type, keyBinding);
	    		} else if (HashSet.class.equals( bindingRequest.clazz.clazz )) {
	       			result = new HashSetBinding(type, keyBinding);
	    		} else {
	   			
	    			if (valueClass==null) throw new BindingConstructionException("Cannot determine key or value class");
	    			ClassBinding valueBindingRequest = new ClassBinding(type.getValueType(), new ClassDataType(valueClass));
	    			Binding valueBinding = unsynchronizedGetBinding(valueBindingRequest, map);
	
	    			if (Map.class.equals( bindingRequest.clazz.clazz ) || TreeMap.class.equals( bindingRequest.clazz.clazz )) {
	    				result = new TreeMapBinding(type, keyBinding, valueBinding);
	    			} else if (HashMap.class.equals( bindingRequest.clazz.clazz )) {
	    				result = new HashMapBinding(type, keyBinding, valueBinding);
	    			} else {
	    				throw new BindingConstructionException("I don't know how to bind "+bindingRequest.clazz.clazz);    			
	    			}
	    			
	    		}
	   			map.put(bindingRequest, result);        	
	   			return result;
	       	} else
	    	if (dataType instanceof OptionalType) {
	    		OptionalType type = (OptionalType) dataType;
	    		OptionalBinding result = new OptionalBindingDefault(type, null);
	    		ClassDataType c = clazz;
	    		Annotation a = c.getAnnotation(Optional.class);
	    		if (a!=null) {
	    			c = c.dropAnnotation(a);
	    		}
	    		ClassBinding componentBindingRequest = new ClassBinding(type.componentType, c);
	        	map.put(bindingRequest, result);
	    		result.componentBinding = unsynchronizedGetBinding(componentBindingRequest, map);
	        	return result;
	    	} else
	    	if (dataType instanceof RecordType) {
	    		RecordType type = (RecordType) dataType;
	    		ReflectionRecordBinding result = new ReflectionRecordBinding(clazz.clazz, type);
	    		map.put(bindingRequest, result);
	    		result.postConstruction(this, map);
	        	return result;
	    	} else
	    	if (dataType instanceof UnionType && clazz.clazz.isEnum()) {
	            UnionType type = (UnionType) dataType;          
	    	    ReflectionEnumBinding result = new ReflectionEnumBinding(type, (Class<Enum<?>>) clazz.clazz );    	    
	            map.put(bindingRequest, result);
	            return result;
	    	} else 
	    	if (dataType instanceof UnionType) {
	    		UnionType type = (UnionType) dataType;    		
	    		ReflectionUnionClassBinding result = new ReflectionUnionClassBinding(type, clazz.clazz);    		
	    		map.put(bindingRequest, result);
	   			result.postConstruction(this, map);
	    		return result;
	    	} else
	    	if (dataType instanceof VariantType && clazz.clazz.equals(Object.class)) {
	    		VariantType type = (VariantType) dataType;
	    		ObjectVariantBinding binding = new ObjectVariantBinding(this);
	    		map.put(bindingRequest, binding);
	    		return binding;
	    	}
    	
		} catch (BindingConstructionException e) {
			map.remove(bindingRequest);
			throw e;
		}
    	
   		throw new BindingConstructionException("Could not figure out a binding for "+bindingRequest);
    }
		
    
    static Annotation[] getFieldAnnotations(Field field) 
    {
    	Class<?> fieldClass = field.getType(); 
    	Annotation[] annotations = field.getAnnotations().clone();
    	Class<?>[] parameterClasses = getParameterClasses(field);
    	
    	if (Set.class.isAssignableFrom(fieldClass) && parameterClasses!=null &&parameterClasses.length==1) {
    		Annotation[] a2 = new Annotation[annotations.length+1];
    		System.arraycopy(annotations, 0, a2, 0, annotations.length);
    		
    		final Class<?> keyType = parameterClasses[0];
    		a2[annotations.length] = new KeyType() {						
				public Class<? extends Annotation> annotationType() { return KeyType.class; }						
				public Class<?> value() { return keyType; }
			};
    		annotations = a2;
    	}    	
    	if (Map.class.isAssignableFrom(fieldClass) && parameterClasses!=null && parameterClasses.length==2) {
    		Annotation[] a2 = new Annotation[annotations.length+2];
    		System.arraycopy(annotations, 0, a2, 0, annotations.length);
    		
    		final Class<?> keyType = parameterClasses[0];
    		final Class<?> valueType = parameterClasses[1];
    		a2[annotations.length] = new KeyType() {						
				public Class<? extends Annotation> annotationType() { return KeyType.class; }						
				public Class<?> value() { return keyType; }
			};
    		a2[annotations.length+1] = new ValueType() {						
				public Class<? extends Annotation> annotationType() { return ValueType.class; }						
				public Class<?> value() { return valueType; }
			};
    		annotations = a2;
    	}    	
    	if (List.class.isAssignableFrom(fieldClass) && parameterClasses!=null && parameterClasses.length==1) {
    		Annotation[] a2 = new Annotation[annotations.length+1];
    		System.arraycopy(annotations, 0, a2, 0, annotations.length);
    		final Class<?> componentType = parameterClasses[0];
    		a2[annotations.length] = new ComponentType() {												
				public Class<? extends Annotation> annotationType() { return ComponentType.class; }
				public Class<?> value() { return componentType; }						
			};
    		annotations = a2;
    	}    	
    	
    	return annotations;
    }
    
    
    /**
     * Get all actual parameters classes, omits generic type definitions
     * 
     * @param f
     * @return
     */
    @SuppressWarnings("unchecked")
	static Class[] getParameterClasses(Field f) {
		Type t = f.getGenericType();
		if (t==null || t instanceof ParameterizedType==false) return new Class[0];		
		ParameterizedType p = (ParameterizedType) t;
		
		int count = 0;
		for (Type x : p.getActualTypeArguments() )
			if (x instanceof Class) count++;
		
		Class[] result = new Class[count];
		int i = 0; 
		for (Type x : p.getActualTypeArguments() )
			if (x instanceof Class)
				result[i++] = (Class) x;
		return result;
    }

    /**
     * Get all actual parameters types, incl. classes and generic types
     * 
     * @param f
     * @return
     */
	static Type[] getParameterTypes(Field f) {
		Type t = f.getGenericType();
		if (t==null || t instanceof ParameterizedType==false) return new Class[0];		
		ParameterizedType p = (ParameterizedType) t;
		return p.getActualTypeArguments();
    }
    
	/**
	 * Get all, including inherited, public, protected and default 
	 * non-transient, non-static, non-final fields. All returned fields have access.
	 * 
	 * @param clazz
	 * @return
	 */
	static Field[] getFields(Class<?> clazz)
	{
		// TODO Sort as order is not guaranteed
		// Sort priority 1-class, 2-modifiers, 3-name, 4-signature
		Field[] fields = getAllFields(clazz);
		
		ArrayList<Field> result = new ArrayList<Field>(fields.length);
		for (Field f : fields) {
			if (Modifier.isStatic(f.getModifiers())) continue;
//			if (Modifier.isFinal(f.getModifiers())) continue;
            if (Modifier.isTransient(f.getModifiers())) continue;           
            if (Modifier.isVolatile(f.getModifiers())) continue;           
			f.setAccessible(true);
			result.add(f);
		}
		return result.toArray( new Field[result.size()] );
	}
	
	// Get all fields except fields of Throwable
	static Field[] getAllFields(Class<?> clazz)
	{
		LinkedList<Class<?>> classes = new LinkedList<Class<?>>();
		while (clazz!=null) {
			classes.addFirst(clazz);
			clazz = clazz.getSuperclass();
			if (clazz==Throwable.class) clazz=null;
		}
		
		ArrayList<Field> result = new ArrayList<Field>();
		for (Class<?> _class : classes) {
			_getAllFields(_class, result);
		}
		
		return result.toArray(new Field[result.size()]);
	}
	
	public static void _getAllFields(Class<?> clazz, Collection<Field> result)
	{
		for (Field m : clazz.getDeclaredFields())
			result.add(m);
	}	    

}


class ClassBinding {
	DataType type; 
	ClassDataType clazz;
	
	
	public ClassBinding(DataType type, ClassDataType clazz) {
		this.type = type;
		this.clazz = clazz;
	}

    @Override
    public boolean equals(Object obj) {
        if (obj == null)
            return false;
        if (!(obj.getClass().equals(this.getClass())))
            return false;
        ClassBinding other = (ClassBinding) obj;
        if (other.type != type && !other.type.equals(type)) return false;
        if (other.clazz != clazz && !other.clazz.equals(clazz)) return false;
        return true;
    }
    
    @Override
    public int hashCode() {
        return type.hashCode() + 31*clazz.hashCode();
    }
    
    @Override
    public String toString() {
        return "<"+type+", "+clazz+">";
    }
}

class ClassDataType {
	Class<?> clazz; 
	Annotation[] annotations;
	Type[] parameters;
	int hash;
	public ClassDataType(Class<?> clazz, Annotation...annotations)
	{
		this.clazz = clazz;
		this.annotations = annotations;
		Annotation[] classAnnotations = clazz.getAnnotations();
		if (classAnnotations!=null && classAnnotations.length>0) {
			if (annotations==null) {
				this.annotations = classAnnotations;
			} else {
				this.annotations = new Annotation[classAnnotations.length + annotations.length];
				System.arraycopy(annotations, 0, this.annotations, 0, annotations.length);
				System.arraycopy(classAnnotations, 0, this.annotations, annotations.length, classAnnotations.length);
			}
		}
		hash = clazz.hashCode();
		if (annotations!=null)
			for (Annotation a : annotations)
				hash = 7*hash + a.hashCode();
	}
	public ClassDataType(Class<?> clazz, Type[] parameters, Annotation...annotations)
	{
		this.clazz = clazz;
		this.annotations = annotations;
		this.parameters = parameters;
		Annotation[] classAnnotations = clazz.getAnnotations();
		if (classAnnotations!=null && classAnnotations.length>0) {
			if (annotations==null) {
				this.annotations = classAnnotations;
			} else {
				this.annotations = new Annotation[classAnnotations.length + annotations.length];
				System.arraycopy(annotations, 0, this.annotations, 0, annotations.length);
				System.arraycopy(classAnnotations, 0, this.annotations, annotations.length, classAnnotations.length);
			}
		}
		hash = clazz.hashCode();
		if (annotations!=null)
			for (Annotation a : annotations)
				hash = 7*hash + a.hashCode();
	}
	
	/**
	 * Get parameters 
	 * @return parameters or null
	 */
	public Type[] getParameters() {
		return parameters;
	}
	
	@SuppressWarnings("unchecked")
	public <A extends Annotation> A getAnnotation(Class<A> annotationClass) 
	{
		if (annotations==null) return null;
		for (Annotation a : annotations)
		{
			if (annotationClass.equals(a.annotationType()))
				return (A) a;
		}
		return null;
	}
	@Override
	public int hashCode() {
		return hash;
	}
	@Override
	public boolean equals(Object obj) {
		if (obj==null) return false;
		if (obj instanceof ClassDataType==false) return false;
		ClassDataType other = (ClassDataType) obj;
		return other.clazz.equals(clazz) &&
			Arrays.deepEquals(annotations, other.annotations);
	}
	
	public ClassDataType dropAnnotation(Annotation a)
	{
		if (annotations==null || annotations.length==0) return this;
		int index = -1;
		for (int i=0; i<annotations.length; i++)
		{
			if (annotations[i]==a) {
				index = i;
				break;
			}
		}
		if (index<0) return this;
		Annotation[] newArray = new Annotation[annotations.length -1];
		if (index>0)
			System.arraycopy(annotations, 0, newArray, 0, index-1);
		if (index<annotations.length-1)
			System.arraycopy(annotations, index+1, newArray, index, annotations.length-index-1);
		return new ClassDataType(clazz, newArray);
	}
	
	@Override
	public String toString() {
		StringBuilder sb = new StringBuilder();
		sb.append(clazz.getName());
		if ( annotations!=null && annotations.length>0 ) {
			sb.append('(');
			for (int i=0; i<annotations.length; i++) {
				Annotation a = annotations[i];
				if (i>0) sb.append(", ");
				sb.append(a);
			}			
			sb.append(')');
		}
		
		return sb.toString();
	}
}

