/*******************************************************************************
 *  Copyright (c) 2010 Association for Decentralized Information Management in
 *  Industry THTH ry.
 *  All rights reserved. This program and the accompanying materials
 *  are made available under the terms of the Eclipse Public License v1.0
 *  which accompanies this distribution, and is available at
 *  http://www.eclipse.org/legal/epl-v10.html
 *
 *  Contributors:
 *      VTT Technical Research Centre of Finland - initial API and implementation
 *******************************************************************************/
package org.simantics.databoard.reflection;

import java.lang.annotation.Annotation;
import java.lang.reflect.Constructor;
import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.util.Map;

import org.simantics.databoard.binding.Binding;
import org.simantics.databoard.binding.RecordBinding;
import org.simantics.databoard.binding.error.BindingConstructionException;
import org.simantics.databoard.binding.error.BindingException;
import org.simantics.databoard.type.DataType;
import org.simantics.databoard.type.RecordType;

/**
 * ReflectionRecordBinding binding of a Record Type and Java Class.
 * If binds Record Values to Java Objects. <p>
 *  
 * 
 * There are three types of classes supported:
 *  1) record-like class 
 *     - All fields are public
 *     - No-argument public constructor
 *     
 *  2) immutable-like class
 *     - All fields are set in constructor
 * 
 *  3) bean-like class
 *     - All fields are set with getter/setter
 */
public class ReflectionRecordBinding extends RecordBinding {

	Constructor<?> argsConstructor, noArgsConstructor;
	Class<?> clazz;
	Field[] fields;
	Method[] getters, setters;

	/**
	 * Create a Reflection Record Bindings. 
	 * 
	 * There is a 2-phase construction.
	 * 
	 * @param clazz
	 * @param type
	 * @throws BindingConstructionException
	 */
	public ReflectionRecordBinding(Class<?> clazz, RecordType type)
	throws BindingConstructionException
	{
		if ( (clazz.getModifiers() & Modifier.ABSTRACT) != 0)
			throw new BindingConstructionException("Cannot create reflection binding to abstract class "+clazz);
		
		try {
			this.clazz = clazz;
			this.type = type;
			
			fields = ReflectionBinding.getFields(clazz);
			for (Field f : fields)
				f.setAccessible(true);

			getters = new Method[fields.length];
			setters = new Method[fields.length];
			for (int i=0; i<fields.length; i++)
			{
				Field f = fields[i];
				String name = f.getName();
				try {
					String getterName = "get"+name.substring(0, 1).toUpperCase()+name.substring(1, name.length());
					Method getter = clazz.getMethod(getterName);
					if (getter.getReturnType().equals(f.getType()))
						getters[i] = getter;
				} catch (NoSuchMethodException e) {
				}
				
				try {
					String setterName = "set"+name.substring(0, 1).toUpperCase()+name.substring(1, name.length());
					Method setter = clazz.getMethod(setterName, fields[i].getType());
					if (setter.getReturnType().equals(void.class))
						setters[i] = setter;					
				} catch (NoSuchMethodException e) {					
				}
			}
			
			// Prepare constuctor for case 2) 
			Class<?>[] constructorArgs = new Class<?>[fields.length];
			for (int i=0; i<fields.length; i++) {
				constructorArgs[i] = fields[i].getType();
			}
			
			try {
				argsConstructor = clazz.getConstructor(constructorArgs);
				argsConstructor.setAccessible(true);
			} catch (NoSuchMethodException e) {}

			try {
				noArgsConstructor = clazz.getConstructor();
				noArgsConstructor.setAccessible(true);				
			} catch (NoSuchMethodException e) {}
			
		} catch (SecurityException e1) {
			throw new BindingConstructionException(e1);
		}
		
       	componentBindings = new Binding[ fields.length ];        	
	}
	
	/**
	 * 2nd Step of the construction. This step creates a binding for each field.
	 * 
	 * @param reflectionScheme
	 * @param map
	 * @throws BindingConstructionException
	 */
	public void postConstruction(ReflectionBinding reflectionScheme, Map<ClassBinding, Binding> map) throws BindingConstructionException 
	{
       	for (int i=0; i<componentBindings.length; i++) {
       		Field f = fields[i];       		
    		Annotation[] annotations = ReflectionBinding.getFieldAnnotations(f);
       		DataType dt = type().getComponentType(i);
       		Class<?> c = f.getType();
    		ClassBinding componentBindingRequest = new ClassBinding(dt, new ClassDataType(c, annotations));
       		componentBindings[i] = reflectionScheme.unsynchronizedGetBinding(componentBindingRequest, map);
       	}			 		
	}
	
	@Override
	public Object createPartial() throws BindingException {
		try {
			if (noArgsConstructor!=null)
				return noArgsConstructor.newInstance();
			else {
				
				// Create default values 
				Object values[] = new Object[fields.length];
				for (int i=0; i<values.length; i++) {
					Binding fb = componentBindings[i];
					values[i] = fb.createDefault();
				}				
				return argsConstructor.newInstance((Object[])values);
//				return clazz.newInstance();
			}
		} catch (InstantiationException e) {
			throw new BindingException(e);
		} catch (IllegalAccessException e) {
			throw new BindingException(e);
		} catch (IllegalArgumentException e) {
			throw new BindingException(e);
		} catch (InvocationTargetException e) {
			throw new BindingException(e);
		}
	}

	@Override
	public Object create(Object... values) 
	throws BindingException {

		// Case 2
		if (argsConstructor != null) 
		{
			try {
				return argsConstructor.newInstance(values);
			} catch (IllegalArgumentException e) {
	            throw new BindingException(e);
			} catch (InstantiationException e) {
	            throw new BindingException(e);
			} catch (IllegalAccessException e) {
	            throw new BindingException(e);
			} catch (InvocationTargetException e) {
	            throw new BindingException(e);
			}
		}

		// Case 1 or 3
		try {
			Object result = null;
			if (noArgsConstructor!=null)
				result = noArgsConstructor.newInstance();
			else
				result = clazz.newInstance();
//			Object result = clazz.newInstance();
			for (int i=0; i<fields.length; i++) {
				Object value = values[i];
				Field f = fields[i];
				Class<?> type = f.getType();
				
				Method setter = setters[i];
				if (setter!=null) {
					// Case 3
					if (type==int.class)
						setter.invoke(result, (Integer)value);
					else if (type==long.class)
						setter.invoke(result, (Long)value);
					else if (type==float.class)
						setter.invoke(result, (Float)value);
					else if (type==double.class)
						setter.invoke(result, (Double)value);
					else if (type==boolean.class)
						setter.invoke(result, (Boolean)value);
					else setter.invoke(result, value);
					
				} else {
					// Case 1
					if (type==int.class)
						f.setInt(result, (Integer)value);
					else if (type==long.class)
						f.setLong(result, (Long)value);
					else if (type==float.class)
						f.setFloat(result, (Float)value);
					else if (type==double.class)
						f.setDouble(result, (Double)value);
					else if (type==boolean.class)
						f.setBoolean(result, (Boolean)value);
					else f.set(result, value);
				}
				
			}
			return result;
		} catch (InstantiationException e) {
            throw new BindingException(e);
		} catch (IllegalAccessException e) {
			throw new BindingException(e);
		} catch (IllegalArgumentException e) {
			throw new BindingException(e);
		} catch (InvocationTargetException e) {
			throw new BindingException(e);
		}
		
	}
	@Override
	public Object getComponent(Object obj, int index) throws BindingException {
		/*
		Field getter = fields[index];		
		if (getter!=null) {
			try {
				return getter.get(obj);
			} catch (IllegalArgumentException e) {
				throw new BindingException(e);
			} catch (IllegalAccessException e) {
				throw new BindingException(e);
			}
		}
		*/
		try {						
			Field f = fields[index];
			Class<?> type = f.getType();
			if (type.isPrimitive()) {
				if (type==int.class) return (Integer) f.getInt(obj);
				else if (type==byte.class) return (Byte) f.getByte(obj);
				else if (type==boolean.class) return (Boolean) f.getBoolean(obj);						
				else if (type==float.class) return (Float) f.getFloat(obj);						
				else if (type==double.class) return (Double) f.getDouble(obj);						
				else if (type==long.class) return (Long) f.getLong(obj);						
				return f.get(obj);
			} else 
			return f.get(obj);
		} catch (IllegalArgumentException e) {
			throw new BindingException(e);
		} catch (IllegalAccessException e) {
			Field getter = fields[index];		
			if (getter!=null) {
				try {
					return getter.get(obj);
				} catch (IllegalArgumentException e2) {
				} catch (IllegalAccessException e2) {
				}
			}
			throw new BindingException(e);
		}
	}

	@Override
	public boolean isInstance(Object obj) {					
		return clazz.isInstance(obj);
	}
	
	@Override
	public boolean isImmutable() {
		return true;
	}
	
	public Class<?> getComponentClass(int index) {
		return fields[index].getDeclaringClass();
	}
	
	@Override
	public void setComponent(Object obj, int index, Object value) throws BindingException {
		try {
			Method setter = setters[index];
			if (setter!=null) {
				setter.invoke(obj, value);
			} else {
				Field f = fields[index];
				Class<?> type = f.getType();
				if (type.isPrimitive()) {
					if (type==int.class) f.setInt(obj, (Integer) value);
					else if (type==byte.class) f.setByte(obj, (Byte) value);
					else if (type==boolean.class) f.setBoolean(obj, (Boolean) value);						
					else if (type==float.class) f.setFloat(obj, (Float) value);						
					else if (type==double.class) f.setDouble(obj, (Double) value);						
					else if (type==long.class) f.setLong(obj, (Long) value);						
					else f.set(obj, value);
				} else {
					f.set(obj, value);
				}
			}
		} catch (IllegalAccessException e) {
			throw new BindingException(e);
		} catch (IllegalArgumentException e) {
			throw new BindingException(e);
		} catch (InvocationTargetException e) {
			e.printStackTrace();
			throw new BindingException(e);
		}					
	}		
	@Override
	public void setComponents(Object obj, Object... value) throws BindingException {
		try {
			for(int i = 0;i<fields.length;++i) {
				Method setter = setters[i];
				if (setter!=null) {
					setter.invoke(obj, value[i]);
				} else {
					Field f = fields[i];
					Class<?> type = f.getType();
					if (type.isPrimitive()) {
						if (type==int.class) f.setInt(obj, (Integer) value[i]);
						else if (type==byte.class) f.setByte(obj, (Byte) value[i]);
						else if (type==boolean.class) f.setBoolean(obj, (Boolean) value[i]);						
						else if (type==float.class) f.setFloat(obj, (Float) value[i]);						
						else if (type==double.class) f.setDouble(obj, (Double) value[i]);						
						else if (type==long.class) f.setLong(obj, (Long) value[i]);						
						else f.set(obj, value[i]);
					} else {
						f.set(obj, value[i]);
					}
				}
			}
		} catch (IllegalAccessException e) {
			throw new BindingException(e);
		} catch (IllegalArgumentException e) {
			throw new BindingException(e);
		} catch (InvocationTargetException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}					
	}	

	
	

}

