/*******************************************************************************
 *  Copyright (c) 2010 Association for Decentralized Information Management in
 *  Industry THTH ry.
 *  All rights reserved. This program and the accompanying materials
 *  are made available under the terms of the Eclipse Public License v1.0
 *  which accompanies this distribution, and is available at
 *  http://www.eclipse.org/legal/epl-v10.html
 *
 *  Contributors:
 *      VTT Technical Research Centre of Finland - initial API and implementation
 *******************************************************************************/
package org.simantics.databoard.util.binary;

import java.io.EOFException;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;

/**
 * This class is a Random Access File implementation to Binary Readable.
 * The implementation is buffered. The implementation is not multi-thread
 * safe. Although file pointer of the backend RandomAccessFile is allowed to 
 * move between read operations of this class. 
 *
 * @author Toni Kalajainen <toni.kalajainen@vtt.fi>
 */
public class FileReadable implements BinaryReadable, Seekable {

	RandomAccessFile file;
	ByteOrder order = ByteOrder.LITTLE_ENDIAN;
	
	/** The internal read buffer */
    byte buf[] = new byte[8192];	
    
    /** The number of valid bytes in the buffer */
    int count;
    
    /** The current position in the buffer */
    int pos;
    
    /** Position of file where buffer was read. */
    long filePosition;
	
	public FileReadable(RandomAccessFile file) throws IOException
	{
		this.file = file;
		filePosition = file.getFilePointer();
	}

	/**
	 * Closes the object. Not this will not close the wrapped random access file.
	 * @throws IOException
	 */
	public synchronized void close() throws IOException {
		buf = null;
	}
	
	public RandomAccessFile getFile() {
		return file;
	}
	
	@Override
	public ByteOrder order() {
		return order;
	}

	@Override
	public void order(ByteOrder order) {
		if (order==null) throw new IllegalArgumentException("null argument");
		this.order = order;
	}

	/**
	 * Get next byte
	 * @return 0..255
	 * @throws IOException
	 */
	int _get() throws IOException
	{
		if (count-pos<1) fill();
		if (count-pos<1) throw new EOFException();
		return buf[pos++] & 0xFF;
	}	
	
	void _get(byte[] dst, int offset, int length) throws IOException {		
		while (length>0) {
			int n = Math.min(count-pos, length);
			System.arraycopy(buf, pos, dst, offset, n);
			offset += n;
			length -= n;
			pos += n;
			
			if (length>0) {
				fill();
				if (count-pos==0) throw new EOFException();
			}
		}
		
	}
	
	private void fill() throws IOException
	{
		// Remove consumed bytes
		if (pos>buf.length/2) {
			System.arraycopy(buf, pos, buf, 0, count-pos);
			count -= pos;
			filePosition += pos;
			pos = 0;			
		}
		
		long bytesLeftInFile = file.length() - filePosition;
		int intBytesLeftInFile = (int) Math.min((long)Integer.MAX_VALUE, bytesLeftInFile);		
		int bytesToFill = Math.min(buf.length - count, intBytesLeftInFile);
		if (bytesToFill==0) return;		

		// Read data
		file.seek(filePosition);
		count += file.read(buf, pos, bytesToFill);
	}
	
	@Override
	public byte get() throws IOException {
		return (byte) _get();
	}

	@Override
	public void get(byte[] dst, int offset, int length) throws IOException {
		_get(dst, offset, length);
	}

	@Override
	public void get(byte[] dst) throws IOException {
		_get(dst, 0, dst.length);
	}

	@Override
	public void get(ByteBuffer buf) throws IOException {
		get(buf, buf.remaining());
	}

	@Override
	public void get(ByteBuffer buf, int length) throws IOException {
		while (length>0) {
			int bytesLeft = count - pos;
			int n = Math.min(buf.remaining(), bytesLeft);
			buf.get(this.buf, pos, n);
			length -= n;
			pos += n;
			fill();			
			if (count - pos == 0)
				throw new EOFException();
		}
	}

	@Override
	public double getDouble() throws IOException {
		return Double.longBitsToDouble(getLong());
	}

	@Override
	public float getFloat() throws IOException {
		return Float.intBitsToFloat(getInt());
	}

	@Override
	public int getInt() throws IOException {
		if (order == ByteOrder.BIG_ENDIAN)
		{		
			return 
				( _get() << 24) |
				( _get() << 16) | 
				( _get() << 8) |
				( _get() );
		} else {
			return 
				( _get() ) |
				( _get() << 8) |
				( _get() << 16) | 
				( _get() << 24);
		}
	}

	@Override
	public long getLong() throws IOException {
		if (order == ByteOrder.BIG_ENDIAN)
		{		
			return
			( ((long)_get()) << 56) |
			( ((long)_get()) << 48) | 
			( ((long)_get()) << 40) |
			( ((long)_get()) << 32) |		
			( ((long)_get()) << 24) |
			( ((long)_get()) << 16) | 
			( ((long)_get()) << 8) |
			( ((long)_get()) );		
		} else {		
			return 
			( ((long)_get())) |
			( ((long)_get()) << 8) |
			( ((long)_get()) << 16) | 
			( ((long)_get()) << 24) |
			( ((long)_get()) << 32) |		
			( ((long)_get()) << 40) |
			( ((long)_get()) << 48) | 
			( ((long)_get()) << 56);
		}	}

	@Override
	public short getShort() throws IOException {
		if (order == ByteOrder.BIG_ENDIAN)
		{
			return (short) ( (_get() << 8) |  _get() ) ;
		} else {
			return (short) ( _get() | (_get() << 8) ) ;
		}
	}

	@Override
	public long length() throws IOException {		
		return file.length();
	}

	@Override
	public long position() throws IOException {
		return filePosition + pos;
	}
	
	public void position(long newPosition) throws IOException {
		if (newPosition>=filePosition && newPosition<filePosition+count) {
			pos = (int) (newPosition - filePosition);
		} else {
			count = 0;
			pos = 0;
			filePosition = newPosition;
		}
	}
	
	/**
	 * Flushes internal buffer
	 */
	public void clearBuffer() {
		count = 0;
	}

	@Override
	public void skip(long bytes) throws IOException {
		position( position() + bytes );
	}
	
}

